!MNH_LIC Copyright 1994-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 22/02/2019: replace Hollerith edit descriptor (deleted from Fortran 95 standard)
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 22/11/2022: modernization of FFts (fixed-format F77 -> free-format F95)
!-----------------------------------------------------------------
MODULE MODE_FFT

USE MODE_MSG

IMPLICIT NONE

PRIVATE

PUBLIC :: FFT991, SET99
PUBLIC :: NVECLEN

INTEGER, PARAMETER :: NVECLEN=1020*16

REAL, PARAMETER :: XSIN36 = 0.587785252292473
REAL, PARAMETER :: XSIN60 = 0.866025403784437
REAL, PARAMETER :: XSIN72 = 0.95105651629515
REAL, PARAMETER :: XQRT5  = 0.559016994374947

CONTAINS

SUBROUTINE SET99( PTRIGS, KFAX, KN )
  ! SUBROUTINE 'SET99' - COMPUTES FACTORS OF KN & TRIGONOMETRIC
  ! FUNCTIONS REQUIRED BY FFT991

  IMPLICIT NONE

  REAL,    DIMENSION(KN), INTENT(INOUT) :: PTRIGS
  INTEGER, DIMENSION(19), INTENT(INOUT) :: KFAX
  INTEGER,                INTENT(IN)    :: KN

  INTEGER, PARAMETER, DIMENSION(7) :: ILFAX = [ 6, 8, 5, 4, 3, 2, 1 ]

  INTEGER :: II, IFAC, IL, IK, IFAX, IU
  INTEGER, DIMENSION(10) :: IJFAX
  REAL :: ZANGLE, ZDEL

  ZDEL=4.0*ASIN(1.0)/REAL(KN)
  DO IK = 0, (KN/2)-1
    ZANGLE=REAL(IK)*ZDEL
    PTRIGS(2*IK+1)=COS(ZANGLE)
    PTRIGS(2*IK+2)=SIN(ZANGLE)
  END DO

  ! FIND FACTORS OF KN (8,6,5,4,3,2; ONLY ONE 8 ALLOWED)
  ! LOOK FOR SIXES FIRST, STORE FACTORS IN DESCENDING ORDER
  IU=KN
  IFAC=6
  IK=0
  IL=1

  DO
    IF ( MOD(IU,IFAC) == 0 ) THEN
      IK=IK+1
      IJFAX(IK)=IFAC
      IF ( IFAC == 8 .AND. IK /= 1 ) THEN
        IJFAX(1)=8
        IJFAX(IK)=6
      END IF
      IU=IU/IFAC
      IF ( IU == 1 ) EXIT
      IF ( IFAC /= 8 ) CYCLE
    END IF

    IL=IL+1
    IFAC=ILFAX(IL)
    IF ( IFAC <= 1 ) CALL PRINT_MSG( NVERB_FATAL, 'GEN', 'SET99', 'illegal factor' )
  END DO

  ! NOW REVERSE ORDER OF FACTORS
  IFAX=IK
  KFAX(1)=IFAX
  DO II=1,IFAX
    KFAX(IFAX+2-II)=IJFAX(II)
  END DO
  KFAX(10)=KN

END SUBROUTINE SET99


SUBROUTINE FFT991( PA, PWORK, PTRIGS, PFAX, KJUMP, KN, KLOT, KSIGN )
  USE MODE_MPPDB

  IMPLICIT NONE

  REAL,    DIMENSION(:),  INTENT(INOUT) :: PA
  REAL,    DIMENSION(:),  INTENT(OUT)   :: PWORK
  REAL,    DIMENSION(:),  INTENT(IN)    :: PTRIGS
  INTEGER, DIMENSION(19), INTENT(IN)    :: PFAX
  INTEGER,                INTENT(IN)    :: KJUMP, KN, KLOT, KSIGN
!
!     SUBROUTINE 'FFT991' - MULTIPLE FAST REAL PERIODIC TRANSFORM
!     SUPERSEDES PREVIOUS ROUTINE 'FFT991'
!
!     REAL TRANSFORM OF LENGTH KN PERFORMED BY REMOVING REDUNDANT
!     OPERATIONS FROM COMPLEX TRANSFORM OF LENGTH KN
!
!     PA IS THE ARRAY CONTAINING INPUT & OUTPUT DATA
!     PWORK IS AN AREA OF SIZE (KN+1)*MIN(KLOT,LOT)
!     PTRIGS IS PA PREVIOUSLY PREPARED LIST OF TRIG FUNCTION VALUES
!     KFAX IS PA PREVIOUSLY PREPARED LIST OF FACTORS OF KN
!     JUMP IS THE INCREMENT BETWEEN THE START OF EACH DATA VECTOR
!     KN IS THE LENGTH OF THE DATA VECTORS
!     KLOT IS THE NUMBER OF DATA VECTORS
!     KSIGN = +1 FOR TRANSFORM FROM SPECTRAL TO GRIDPOINT
!           = -1 FOR TRANSFORM FROM GRIDPOINT TO SPECTRAL
!
!     ORDERING OF COEFFICIENTS:
!         PA(0),PB(0),PA(1),PB(1),PA(2),PB(2),...,PA(KN/2),PB(KN/2)
!         WHERE PB(0)=PB(KN/2)=0; (KN+2) LOCATIONS REQUIRED
!
!     ORDERING OF DATA:
!         X(0),X(1),X(2),...,X(KN-1), 0 , 0 ; (KN+2) LOCATIONS REQUIRED
!
!     VECTORIZATION IS ACHIEVED ON CRAY BY DOING THE TRANSFORMS
!     IN PARALLEL
!
!     KN MUST BE COMPOSED OF FACTORS 2,3 & 5 BUT DOES NOT HAVE TO BE EVEN
!
!     DEFINITION OF TRANSFORMS:
!     -------------------------
!
!     KSIGN=+1: X(J)=SUM(K=0,...,KN-1)(PC(K)*EXP(2*I*J*K*PI/KN))
!         WHERE PC(K)=PA(K)+I*PB(K) AND PC(KN-K)=PA(K)-I*PB(K)
!
!     KSIGN=-1: PA(K)=(1/KN)*SUM(J=0,...,KN-1)(X(J)*COS(2*J*K*PI/KN))
!               PB(K)=-(1/KN)*SUM(J=0,...,KN-1)(X(J)*SIN(2*J*K*PI/KN))
!
  INTEGER :: IFAX, INX, INBLOX, INVEX
  INTEGER :: I, J, IA, ISTART, ILA, IGO
  INTEGER :: INB, IK, IFAC, IERR
  INTEGER :: IIBASE, IJBASE
  INTEGER :: II, JJ, IX, IZ
  INTEGER :: I0

  IF (MPPDB_INITIALIZED) THEN
    !Check all IN arrays
    CALL MPPDB_CHECK( PTRIGS, "FFT991 beg:PTRIGS" )
    CALL MPPDB_CHECK( PFAX, "FFT991 beg:KFAX" )
    !Check all INOUT arrays
    CALL MPPDB_CHECK( PA, "FFT991 beg:PA" )
  END IF

!$acc data present( PA, PWORK )

  ! Initialisation of PWORK useful to compare results with MPPDB_CHECK (otherwise all values are not set by FFT991
  PWORK(:) = 0.

#if 0
  !PW: Original version: but in that case, intent of PTRIGS and KFAX are not correct
  IF(PFAX(10).NE.KN) CALL SET99(PTRIGS,KFAX,KN)
#else
  IF(PFAX(10).NE.KN) CALL PRINT_MSG( NVERB_FATAL, 'GEN', 'FFT991', 'PTRIGS and KFAX not initialised (set99 not yet called)' )
#endif
  IFAX=PFAX(1)
  INX=KN+1
  IF (MOD(KN,2).EQ.1) INX=KN
  INBLOX=1+(KLOT-1)/NVECLEN
  INVEX=KLOT-(INBLOX-1)*NVECLEN

  IF ( KSIGN == 1 ) THEN
!
!   KSIGN=+1, SPECTRAL TO GRIDPOINT TRANSFORM
!   -----------------------------------------
    ISTART=1
    DO INB=1,INBLOX
      IA=ISTART
!$acc kernels
!CDIR NODEP
!*vocl loop,novrec
!$acc loop independent private( I )
      DO J=1,INVEX
        I = ISTART + ( J - 1 ) * KJUMP
        PA(I+1)=0.5*PA(I)
      END DO
!$acc end kernels
      IF ( MOD(KN,2) == 0 ) THEN
!$acc kernels
        I0 = ISTART + KN
!CDIR NODEP
!*vocl loop,novrec
!$acc loop independent private( I )
        DO J=1,INVEX
          I = I0 + ( J - 1 ) * KJUMP
          PA(I)=0.5*PA(I)
        END DO
!$acc end kernels
      END IF
      IA=ISTART+1
      ILA=1
      IGO=+1
!
      DO IK=1,IFAX
        IFAC=PFAX(IK+1)
        IERR=-1
        IF ( IGO == 1 ) THEN
          CALL RPASSM( PA(IA:), PA(IA+ILA:), PWORK(:), PWORK(IFAC*ILA+1:), PTRIGS(:), &
                       KJUMP, INX, INVEX, KN, IFAC, ILA, IERR )
        ELSE
          CALL RPASSM( PWORK(:), PWORK(ILA+1:), PA(IA:), PA(IA+IFAC*ILA:), PTRIGS(:), &
                       INX, KJUMP, INVEX, KN, IFAC, ILA, IERR )
        END IF
        IF (IERR.NE.0) GO TO 500
        ILA=IFAC*ILA
        IGO=-IGO
        IA=ISTART
      END DO
!
!     IF NECESSARY, COPY RESULTS BACK TO PA
!     -------------------------------------
      IF ( MOD(IFAX,2) == 1 ) THEN
!$acc kernels
!CDIR NODEP
!*vocl loop,novrec
!$acc loop independent private( IIBASE, IJBASE )
        DO JJ = 1, INVEX
          IIBASE = 1  + (JJ-1) * INX
          IJBASE = IA + (JJ-1) * KJUMP
          PA(IJBASE:IJBASE+KN-1) = PWORK(IIBASE:IIBASE+KN-1)
        END DO
!$acc end kernels
      END IF
!
!     FILL IN ZEROS AT END
!     --------------------
!$acc kernels
      PA(ISTART+KN::KJUMP) = 0.0
      PA(ISTART+KN+1::KJUMP) = 0.0
!$acc end kernels
!
      ISTART=ISTART+INVEX*KJUMP
      INVEX=NVECLEN
    END DO

  ELSE IF ( KSIGN == -1 ) THEN
!
!     KSIGN=-1, GRIDPOINT TO SPECTRAL TRANSFORM
!     -----------------------------------------
    ISTART=1
    DO INB=1,INBLOX
      IA=ISTART
      ILA=KN
      IGO=+1
!
      DO IK=1,IFAX
        IFAC=PFAX(IFAX+2-IK)
        ILA=ILA/IFAC
        IERR=-1
        IF ( IGO == 1 ) THEN
          CALL QPASSM( PA(IA:), PA(IA+IFAC*ILA:), PWORK(:), PWORK(ILA+1:), PTRIGS(:), &
                       KJUMP, INX, INVEX, KN, IFAC, ILA, IERR )
        ELSE
          CALL QPASSM( PWORK(:), PWORK(IFAC*ILA+1:), PA(IA:), PA(IA+ILA:), PTRIGS(:), &
                       INX, KJUMP, INVEX, KN, IFAC, ILA, IERR )
        END IF
        IF (IERR.NE.0) GO TO 500
        IGO=-IGO
        IA=ISTART+1
      END DO
!
!     IF NECESSARY, COPY RESULTS BACK TO PA
!     -------------------------------------
      IF ( MOD(IFAX,2) == 1 ) THEN
!$acc kernels
        !CDIR NODEP
!*vocl loop,novrec
!$acc loop independent private( IIBASE, IJBASE )
        DO JJ = 1, INVEX
          IIBASE = 1  + (JJ-1) * INX
          IJBASE = IA + (JJ-1) * KJUMP
          PA(IJBASE:IJBASE+KN-1) = PWORK(IIBASE:IIBASE+KN-1)
        END DO
!$acc end kernels
      END IF
!
!     SHIFT PA(0) & FILL IN ZERO IMAG PARTS
!     -------------------------------------
!$acc kernels
!CDIR NODEP
!*vocl loop,novrec
!$acc loop independent private( I )
      DO J=1,INVEX
        IX  = ISTART + ( J - 1 ) * KJUMP
        PA(IX)=PA(IX+1)
        PA(IX+1)=0.0
      END DO
!$acc end kernels
      IF ( MOD(KN,2) == 0 ) THEN
!$acc kernels
        PA(ISTART+KN+1::KJUMP) = 0.0
!$acc end kernels
      END IF
!
      ISTART=ISTART+INVEX*KJUMP
      INVEX=NVECLEN
    END DO
  END IF
!
!     ERROR MESSAGES
!     --------------
  500 CONTINUE
      IF ( IERR /= 0 ) THEN
      IF(IERR.EQ.1) THEN
  520 FORMAT('VECTOR LENGTH =',I4,', GREATER THAN NVECLEN')
      ELSEIF(IERR.EQ.2) THEN
      WRITE(6,540) IFAC
  540 FORMAT( 'FACTOR =',I3,', NOT CATERED FOR')
      ELSE
      WRITE(6,560) IFAC
  560 FORMAT('FACTOR =',I3,', ONLY CATERED FOR IF ILA*IFAC=KN')
      ENDIF
      END IF

!$acc end data

  IF (MPPDB_INITIALIZED) THEN
    !Check all INOUT arrays
    CALL MPPDB_CHECK( PA, "FFT991 end:PA" )
    !Check all OUT arrays
    CALL MPPDB_CHECK( PWORK, "FFT991 end:PWORK" )
  END IF
END SUBROUTINE FFT991



SUBROUTINE RPASSM( PA, PB, PC, PD, PTRIGS, KINC3, KINC4, KLOT, KN, KFAC, KLA, KERR )

  IMPLICIT NONE

  REAL, DIMENSION(:), INTENT(IN)    :: PA
  REAL, DIMENSION(:), INTENT(IN)    :: PB
  REAL, DIMENSION(:), INTENT(INOUT) :: PC
  REAL, DIMENSION(:), INTENT(INOUT) :: PD
  REAL, DIMENSION(:), INTENT(IN)    :: PTRIGS
  INTEGER,            INTENT(IN)    :: KINC3, KINC4, KLOT, KN, KFAC, KLA
  INTEGER,            INTENT(OUT)   :: KERR
!
!     SUBROUTINE 'RPASSM' - PERFORMS ONE PASS THROUGH DATA AS PART
!     OF MULTIPLE REAL FFT (FOURIER SYNTHESIS) ROUTINE
!
!     KA IS FIRST REAL INPUT VECTOR
!         EQUIVALENCE PB(1) WITH PA (KLA+1)
!     PC IS FIRST REAL OUTPUT VECTOR
!         EQUIVALENCE PD(1) WITH PC(KFAC*KLA+1)
!     PTRIGS IS PA PRECALCULATED LIST OF SINES & COSINES
!     KINC3 IS THE INCREMENT BETWEEN INPUT VECTORS PA
!     KINC4 IS THE INCREMENT BETWEEN OUTPUT VECTORS PC
!     KLOT IS THE NUMBER OF VECTORS
!     KN IS THE LENGTH OF THE VECTORS
!     KFAC IS THE CURRENT FACTOR OF KN
!     KLA IS THE PRODUCT OF PREVIOUS FACTORS
!     KERR IS AN ERROR INDICATOR:
!              0 - PASS COMPLETED WITHOUT ERROR
!              1 - KLOT GREATER THAN NVECLEN
!              2 - KFAC NOT CATERED FOR
!              3 - KFAC ONLY CATERED FOR IF KLA=KN/KFAC
!
!-----------------------------------------------------------------------
!
  INTEGER :: IM, IINK, JINK, JUMP, ISTOP
  INTEGER :: IBAD, IIBASE, IJBASE, IGO
  INTEGER :: II, IJ, IIA, IIB, IJA, IJB
  INTEGER :: IIL, IJK, IIK, IKB, IIC, IJC, IKC
  INTEGER :: IID, IJD, IKD, IIE, IJE, IKE, IIF, IJF, IKF, IJG, IJH
  INTEGER :: IIA0, IIB0, IIC0, IID0, IIE0, IIF0, IJBASE0
  REAL :: ZA10, ZA11, ZA20, ZA21, ZB10, ZB11, ZB20, ZB21
  REAL :: ZC1, ZC2, ZC3, ZC4, ZC5, ZS1, ZS2, ZS3, ZS4, ZS5
  REAL :: ZQQRT5, ZSIN45, ZSSIN36, ZSSIN45, ZSSIN60, ZSSIN72


!$acc data present( PA, PB, PC, PD ) copyin( PTRIGS )

!acc kernels
      IM=KN/KFAC
      IINK=KLA
      JINK=KLA
      JUMP=(KFAC-1)*JINK
      ISTOP=(KN-KFAC)/(2*KFAC)

!acc end kernels
      IBAD=1
      IF (KLOT.GT.NVECLEN) GO TO 910
      IIBASE=0
      IJBASE=0
      IGO=KFAC-1
      IF (IGO.EQ.7) IGO=6
      IBAD=2
      IF (IGO.LT.1.OR.IGO.GT.6) GO TO 910

      SELECT CASE ( KFAC )

      CASE ( 2 )
!
!     CODING FOR FACTOR 2
!     -------------------
      IIA=1
      IIB=IIA+2*IM-KLA
      IJA=1
      IJB=IJA+JINK

      IF ( KLA /= IM ) THEN

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=PA(IIA+II)+PA(IIB+II)
            PC(IJB+IJ)=PA(IIA+II)-PA(IIB+II)
          END DO
        END DO
        IIBASE=IIBASE+KLA
        IJBASE=IJBASE+KLA
!$acc end kernels
        IIA=IIA+IINK
        IINK=2*IINK
        IIB=IIB-IINK
        IIBASE=0
        IJBASE=IJBASE+JUMP
        JUMP=2*JUMP+JINK
        IF ( IIA /= IIB ) THEN
!$acc kernels
          IJBASE0 = IJBASE
          IIA0 = IIA
          IIB0 = IIB
!$acc loop independent private( IKB, ZC1, ZS1, IJBASE, IIA, IIB )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            IJBASE = IJBASE0 + (IIK-KLA)/KLA * (KLA + JUMP )
            IIA = IIA0 + (IIK-KLA)/KLA * IINK
            IIB = IIB0 - (IIK-KLA)/KLA * IINK
!$acc loop independent
            DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
              DO IJK=1,KLOT
                II =         IIL - 1 + (IJK - 1 ) * KINC3
                IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
                PC(IJA+IJ)=PA(IIA+II)+PA(IIB+II)
                PD(IJA+IJ)=PB(IIA+II)-PB(IIB+II)
                PC(IJB+IJ)=ZC1*(PA(IIA+II)-PA(IIB+II))-ZS1*(PB(IIA+II)+PB(IIB+II))
                PD(IJB+IJ)=ZS1*(PA(IIA+II)-PA(IIB+II))+ZC1*(PB(IIA+II)+PB(IIB+II))
              END DO
            END DO
          END DO
          IJBASE = IJBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + JUMP )
          IIA = IIA0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIB = IIB0 - ((ISTOP-KLA)/KLA+1) * IINK
!$acc end kernels
        END IF
        IF ( IIA <= IIB ) THEN
!$acc kernels
          IIBASE=0
!$acc loop independent
          DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIA+II)
              PC(IJB+IJ)=-PB(IIA+II)
            END DO
          END DO
          IIBASE=IIBASE+KLA
          IJBASE=IJBASE+KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=2.0*(PA(IIA+II)+PA(IIB+II))
            PC(IJB+IJ)=2.0*(PA(IIA+II)-PA(IIB+II))
          END DO
        END DO
        IIBASE=IIBASE+KLA
        IJBASE=IJBASE+KLA
!$acc end kernels
      END IF
      IBAD=0


      CASE ( 3 )
!
!     CODING FOR FACTOR 3
!     -------------------
      IIA=1
      IIB=IIA+2*IM-KLA
      IIC=IIB
      IJA=1
      IJB=IJA+JINK
      IJC=IJB+JINK

      IF ( KLA /= IM ) THEN

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=PA(IIA+II)+PA(IIB+II)
            PC(IJB+IJ)=(PA(IIA+II)-0.5*PA(IIB+II))-(XSIN60*(PB(IIB+II)))
            PC(IJC+IJ)=(PA(IIA+II)-0.5*PA(IIB+II))+(XSIN60*(PB(IIB+II)))
          END DO
        END DO
        IIBASE=IIBASE+KLA
        IJBASE=IJBASE+KLA
!$acc end kernels
        IIA=IIA+IINK
        IINK=2*IINK
        IIB=IIB+IINK
        IIC=IIC-IINK
        IJBASE=IJBASE+JUMP
        JUMP=2*JUMP+JINK
        IF ( IIA /= IIC ) THEN
!$acc kernels
          IJBASE0 = IJBASE
          IIA0 = IIA
          IIB0 = IIB
          IIC0 = IIC
!$acc loop independent private( IKB, IKC, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, IJBASE, IIA, IIB, IIC )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            IJBASE = IJBASE0 + (IIK-KLA)/KLA * (KLA + JUMP )
            IIA = IIA0 + (IIK-KLA)/KLA * IINK
            IIB = IIB0 + (IIK-KLA)/KLA * IINK
            IIC = IIC0 - (IIK-KLA)/KLA * IINK
!$acc loop independent
            DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
              DO IJK=1,KLOT
                II =         IIL - 1 + (IJK - 1 ) * KINC3
                IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
                PC(IJA+IJ)=PA(IIA+II)+(PA(IIB+II)+PA(IIC+II))
                PD(IJA+IJ)=PB(IIA+II)+(PB(IIB+II)-PB(IIC+II))
                PC(IJB+IJ)=                                                      &
                    ZC1*((PA(IIA+II)-0.5*(PA(IIB+II)+PA(IIC+II)))-(XSIN60*(PB(IIB+II)+PB(IIC+ &
                  II))))                                                       &
                  -ZS1*((PB(IIA+II)-0.5*(PB(IIB+II)-PB(IIC+II)))+(XSIN60*(PA(IIB+II)-PA(IIC+ &
                  II))))
                PD(IJB+IJ)=                                                      &
                    ZS1*((PA(IIA+II)-0.5*(PA(IIB+II)+PA(IIC+II)))-(XSIN60*(PB(IIB+II)+PB(IIC+ &
                  II))))                                                       &
                  +ZC1*((PB(IIA+II)-0.5*(PB(IIB+II)-PB(IIC+II)))+(XSIN60*(PA(IIB+II)-PA(IIC+ &
                  II))))
                PC(IJC+IJ)=                                                      &
                    ZC2*((PA(IIA+II)-0.5*(PA(IIB+II)+PA(IIC+II)))+(XSIN60*(PB(IIB+II)+PB(IIC+ &
                  II))))                                                       &
                  -ZS2*((PB(IIA+II)-0.5*(PB(IIB+II)-PB(IIC+II)))-(XSIN60*(PA(IIB+II)-PA(IIC+ &
                  II))))
                PD(IJC+IJ)=                                                      &
                    ZS2*((PA(IIA+II)-0.5*(PA(IIB+II)+PA(IIC+II)))+(XSIN60*(PB(IIB+II)+PB(IIC+ &
                  II))))                                                       &
                  +ZC2*((PB(IIA+II)-0.5*(PB(IIB+II)-PB(IIC+II)))-(XSIN60*(PA(IIB+II)-PA(IIC+ &
                  II))))
              END DO
            END DO
          END DO
          IJBASE = IJBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + JUMP )
          IIA = IIA0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIB = IIB0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIC = IIC0 - ((ISTOP-KLA)/KLA+1) * IINK
!$acc end kernels
        END IF

        IF ( IIA <= IIC ) THEN
!$acc kernels
          IIBASE=0
!$acc loop independent
          DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIA+II)+PA(IIB+II)
              PC(IJB+IJ)=(0.5*PA(IIA+II)-PA(IIB+II))-(XSIN60*PB(IIA+II))
              PC(IJC+IJ)=-(0.5*PA(IIA+II)-PA(IIB+II))-(XSIN60*PB(IIA+II))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
        ZSSIN60=2.0*XSIN60
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=2.0*(PA(IIA+II)+PA(IIB+II))
            PC(IJB+IJ)=(2.0*PA(IIA+II)-PA(IIB+II))-(ZSSIN60*PB(IIB+II))
            PC(IJC+IJ)=(2.0*PA(IIA+II)-PA(IIB+II))+(ZSSIN60*PB(IIB+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF

      IBAD=0


      CASE ( 4 )
!
!     CODING FOR FACTOR 4
!     -------------------
      IIA=1
      IIB=IIA+2*IM-KLA
      IIC=IIB+2*IM
      IID=IIB
      IJA=1
      IJB=IJA+JINK
      IJC=IJB+JINK
      IJD=IJC+JINK

      IF ( KLA /=  IM) THEN

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=(PA(IIA+II)+PA(IIC+II))+PA(IIB+II)
            PC(IJB+IJ)=(PA(IIA+II)-PA(IIC+II))-PB(IIB+II)
            PC(IJC+IJ)=(PA(IIA+II)+PA(IIC+II))-PA(IIB+II)
            PC(IJD+IJ)=(PA(IIA+II)-PA(IIC+II))+PB(IIB+II)
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IIA=IIA+IINK
        IINK=2*IINK
        IIB=IIB+IINK
        IIC=IIC-IINK
        IID=IID-IINK
        IJBASE=IJBASE+JUMP
        JUMP=2*JUMP+JINK
        IF ( IIB /=  IIC ) THEN
!$acc kernels
          IJBASE0 = IJBASE
          IIA0 = IIA
          IIB0 = IIB
          IIC0 = IIC
          IID0 = IID
!$acc loop independent private( IKB, IKC, IKD, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, IJBASE, IIA, IIB, IIC, IID )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            IKD=IKC+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            ZC3=PTRIGS(IKD+1)
            ZS3=PTRIGS(IKD+2)
            IJBASE = IJBASE0 + (IIK-KLA)/KLA * (KLA + JUMP )
            IIA = IIA0 + (IIK-KLA)/KLA * IINK
            IIB = IIB0 + (IIK-KLA)/KLA * IINK
            IIC = IIC0 - (IIK-KLA)/KLA * IINK
            IID = IID0 - (IIK-KLA)/KLA * IINK
!$acc loop independent
            DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
              DO IJK=1,KLOT
                II =         IIL - 1 + (IJK - 1 ) * KINC3
                IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
                PC(IJA+IJ)=(PA(IIA+II)+PA(IIC+II))+(PA(IIB+II)+PA(IID+II))
                PD(IJA+IJ)=(PB(IIA+II)-PB(IIC+II))+(PB(IIB+II)-PB(IID+II))
                PC(IJC+IJ)=                                     &
                    ZC2*((PA(IIA+II)+PA(IIC+II))-(PA(IIB+II)+PA(IID+II))) &
                   -ZS2*((PB(IIA+II)-PB(IIC+II))-(PB(IIB+II)-PB(IID+II)))
                PD(IJC+IJ)=                                     &
                    ZS2*((PA(IIA+II)+PA(IIC+II))-(PA(IIB+II)+PA(IID+II))) &
                   +ZC2*((PB(IIA+II)-PB(IIC+II))-(PB(IIB+II)-PB(IID+II)))
                PC(IJB+IJ)=                                     &
                    ZC1*((PA(IIA+II)-PA(IIC+II))-(PB(IIB+II)+PB(IID+II))) &
                   -ZS1*((PB(IIA+II)+PB(IIC+II))+(PA(IIB+II)-PA(IID+II)))
                PD(IJB+IJ)=                                     &
                    ZS1*((PA(IIA+II)-PA(IIC+II))-(PB(IIB+II)+PB(IID+II))) &
                   +ZC1*((PB(IIA+II)+PB(IIC+II))+(PA(IIB+II)-PA(IID+II)))
                PC(IJD+IJ)=                                     &
                    ZC3*((PA(IIA+II)-PA(IIC+II))+(PB(IIB+II)+PB(IID+II))) &
                   -ZS3*((PB(IIA+II)+PB(IIC+II))-(PA(IIB+II)-PA(IID+II)))
                PD(IJD+IJ)=                                     &
                    ZS3*((PA(IIA+II)-PA(IIC+II))+(PB(IIB+II)+PB(IID+II))) &
                   +ZC3*((PB(IIA+II)+PB(IIC+II))-(PA(IIB+II)-PA(IID+II)))
              END DO
            END DO
          END DO
          IJBASE = IJBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + JUMP )
          IIA = IIA0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIB = IIB0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIC = IIC0 - ((ISTOP-KLA)/KLA+1) * IINK
          IID = IID0 - ((ISTOP-KLA)/KLA+1) * IINK
!$acc end kernels
        END IF

        IF ( IIB <= IIC ) THEN
!$acc kernels
          IIBASE=0
          ZSIN45=SQRT(0.5)
!$acc loop independent
          DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIA+II)+PA(IIB+II)
              PC(IJB+IJ)=ZSIN45*((PA(IIA+II)-PA(IIB+II))-(PB(IIA+II)+PB(IIB+II)))
              PC(IJC+IJ)=PB(IIB+II)-PB(IIA+II)
              PC(IJD+IJ)=-ZSIN45*((PA(IIA+II)-PA(IIB+II))+(PB(IIA+II)+PB(IIB+II)))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE ! KLA == M

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=2.0*((PA(IIA+II)+PA(IIC+II))+PA(IIB+II))
            PC(IJB+IJ)=2.0*((PA(IIA+II)-PA(IIC+II))-PB(IIB+II))
            PC(IJC+IJ)=2.0*((PA(IIA+II)+PA(IIC+II))-PA(IIB+II))
            PC(IJD+IJ)=2.0*((PA(IIA+II)-PA(IIC+II))+PB(IIB+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF

      IBAD=0


      CASE ( 5 )
!
!     CODING FOR FACTOR 5
!     -------------------
      IIA=1
      IIB=IIA+2*IM-KLA
      IIC=IIB+2*IM
      IID=IIC
      IIE=IIB
      IJA=1
      IJB=IJA+JINK
      IJC=IJB+JINK
      IJD=IJC+JINK
      IJE=IJD+JINK

      IF ( KLA /=  IM ) THEN

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=PA(IIA+II)+(PA(IIB+II)+PA(IIC+II))
            PC(IJB+IJ)=((PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))+XQRT5*(PA(IIB+II)-PA(IIC+ &
              II)))                                                        &
                -(XSIN72*PB(IIB+II)+XSIN36*PB(IIC+II))
            PC(IJC+IJ)=((PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))-XQRT5*(PA(IIB+II)-PA(IIC+ &
              II)))                                                        &
                -(XSIN36*PB(IIB+II)-XSIN72*PB(IIC+II))
            PC(IJD+IJ)=((PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))-XQRT5*(PA(IIB+II)-PA(IIC+ &
              II)))                                                        &
                +(XSIN36*PB(IIB+II)-XSIN72*PB(IIC+II))
            PC(IJE+IJ)=((PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))+XQRT5*(PA(IIB+II)-PA(IIC+ &
              II)))                                                        &
                +(XSIN72*PB(IIB+II)+XSIN36*PB(IIC+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IIA=IIA+IINK
        IINK=2*IINK
        IIB=IIB+IINK
        IIC=IIC+IINK
        IID=IID-IINK
        IIE=IIE-IINK
        IJBASE=IJBASE+JUMP
        JUMP=2*JUMP+JINK
        IF ( IIB /= IID ) THEN
!$acc kernels
          IJBASE0 = IJBASE
          IIA0 = IIA
          IIB0 = IIB
          IIC0 = IIC
          IID0 = IID
          IIE0 = IIE
!$acc loop independent private( IKB, IKC, IKD, IKE, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, ZC4, ZS4, IJBASE, IIA, IIB, IIC, IID, IIE )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            IKD=IKC+IKB
            IKE=IKD+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            ZC3=PTRIGS(IKD+1)
            ZS3=PTRIGS(IKD+2)
            ZC4=PTRIGS(IKE+1)
            ZS4=PTRIGS(IKE+2)
            IJBASE = IJBASE0 + (IIK-KLA)/KLA * (KLA + JUMP )
            IIA = IIA0 + (IIK-KLA)/KLA * IINK
            IIB = IIB0 + (IIK-KLA)/KLA * IINK
            IIC = IIC0 + (IIK-KLA)/KLA * IINK
            IID = IID0 - (IIK-KLA)/KLA * IINK
            IIE = IIE0 - (IIK-KLA)/KLA * IINK
!$acc loop independent
            DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA10, ZA11, ZA20, ZA21, ZB10, ZB11, ZB20, ZB21 )
              DO IJK=1,KLOT
                II =         IIL - 1 + (IJK - 1 ) * KINC3
                IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4

                ZA10=(PA(IIA+II)-0.25*((PA(IIB+II)+PA(IIE+II))+(PA(IIC+II)+PA(IID+II)))) &
                    +XQRT5*((PA(IIB+II)+PA(IIE+II))-(PA(IIC+II)+PA(IID+II)))
                ZA20=(PA(IIA+II)-0.25*((PA(IIB+II)+PA(IIE+II))+(PA(IIC+II)+PA(IID+II)))) &
                    -XQRT5*((PA(IIB+II)+PA(IIE+II))-(PA(IIC+II)+PA(IID+II)))
                ZB10=(PB(IIA+II)-0.25*((PB(IIB+II)-PB(IIE+II))+(PB(IIC+II)-PB(IID+II)))) &
                    +XQRT5*((PB(IIB+II)-PB(IIE+II))-(PB(IIC+II)-PB(IID+II)))
                ZB20=(PB(IIA+II)-0.25*((PB(IIB+II)-PB(IIE+II))+(PB(IIC+II)-PB(IID+II)))) &
                    -XQRT5*((PB(IIB+II)-PB(IIE+II))-(PB(IIC+II)-PB(IID+II)))
                ZA11=XSIN72*(PB(IIB+II)+PB(IIE+II))+XSIN36*(PB(IIC+II)+PB(IID+II))
                ZA21=XSIN36*(PB(IIB+II)+PB(IIE+II))-XSIN72*(PB(IIC+II)+PB(IID+II))
                ZB11=XSIN72*(PA(IIB+II)-PA(IIE+II))+XSIN36*(PA(IIC+II)-PA(IID+II))
                ZB21=XSIN36*(PA(IIB+II)-PA(IIE+II))-XSIN72*(PA(IIC+II)-PA(IID+II))

                PC(IJA+IJ)=PA(IIA+II)+((PA(IIB+II)+PA(IIE+II))+(PA(IIC+II)+PA(IID+II)))
                PD(IJA+IJ)=PB(IIA+II)+((PB(IIB+II)-PB(IIE+II))+(PB(IIC+II)-PB(IID+II)))
                PC(IJB+IJ)=ZC1*(ZA10-ZA11)-ZS1*(ZB10+ZB11)
                PD(IJB+IJ)=ZS1*(ZA10-ZA11)+ZC1*(ZB10+ZB11)
                PC(IJE+IJ)=ZC4*(ZA10+ZA11)-ZS4*(ZB10-ZB11)
                PD(IJE+IJ)=ZS4*(ZA10+ZA11)+ZC4*(ZB10-ZB11)
                PC(IJC+IJ)=ZC2*(ZA20-ZA21)-ZS2*(ZB20+ZB21)
                PD(IJC+IJ)=ZS2*(ZA20-ZA21)+ZC2*(ZB20+ZB21)
                PC(IJD+IJ)=ZC3*(ZA20+ZA21)-ZS3*(ZB20-ZB21)
                PD(IJD+IJ)=ZS3*(ZA20+ZA21)+ZC3*(ZB20-ZB21)
              END DO
            END DO
          END DO
          IJBASE = IJBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + JUMP )
          IIA = IIA0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIB = IIB0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIC = IIC0 + ((ISTOP-KLA)/KLA+1) * IINK
          IID = IID0 - ((ISTOP-KLA)/KLA+1) * IINK
          IIE = IIE0 - ((ISTOP-KLA)/KLA+1) * IINK
!$acc end kernels
        END IF

        IF ( IIB <= IID ) THEN
!$acc kernels
          IIBASE=0
!$acc loop independent
          DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=(PA(IIA+II)+PA(IIB+II))+PA(IIC+II)
              PC(IJB+IJ)=(XQRT5*(PA(IIA+II)-PA(IIB+II))+(0.25*(PA(IIA+II)+PA(IIB+II))-PA(IIC+ &
                II)))                                                        &
                  -(XSIN36*PB(IIA+II)+XSIN72*PB(IIB+II))
              PC(IJE+IJ)=-(XQRT5*(PA(IIA+II)-PA(IIB+II))+(0.25*(PA(IIA+II)+PA(IIB+II))-PA(IIC+ &
                II)))                                                         &
                  -(XSIN36*PB(IIA+II)+XSIN72*PB(IIB+II))
              PC(IJC+IJ)=(XQRT5*(PA(IIA+II)-PA(IIB+II))-(0.25*(PA(IIA+II)+PA(IIB+II))-PA(IIC+ &
                II)))                                                        &
                  -(XSIN72*PB(IIA+II)-XSIN36*PB(IIB+II))
              PC(IJD+IJ)=-(XQRT5*(PA(IIA+II)-PA(IIB+II))-(0.25*(PA(IIA+II)+PA(IIB+II))-PA(IIC+ &
                II)))                                                         &
                  -(XSIN72*PB(IIA+II)-XSIN36*PB(IIB+II))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
        ZQQRT5=2.0*XQRT5
        ZSSIN36=2.0*XSIN36
        ZSSIN72=2.0*XSIN72
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=2.0*(PA(IIA+II)+(PA(IIB+II)+PA(IIC+II)))
            PC(IJB+IJ)=(2.0*(PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))                 &
                +ZQQRT5*(PA(IIB+II)-PA(IIC+II)))-(ZSSIN72*PB(IIB+II)+ZSSIN36*PB(IIC+II))
            PC(IJC+IJ)=(2.0*(PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))                 &
                -ZQQRT5*(PA(IIB+II)-PA(IIC+II)))-(ZSSIN36*PB(IIB+II)-ZSSIN72*PB(IIC+II))
            PC(IJD+IJ)=(2.0*(PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))                 &
                -ZQQRT5*(PA(IIB+II)-PA(IIC+II)))+(ZSSIN36*PB(IIB+II)-ZSSIN72*PB(IIC+II))
            PC(IJE+IJ)=(2.0*(PA(IIA+II)-0.25*(PA(IIB+II)+PA(IIC+II)))                 &
                +ZQQRT5*(PA(IIB+II)-PA(IIC+II)))+(ZSSIN72*PB(IIB+II)+ZSSIN36*PB(IIC+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF

      IBAD=0


      CASE ( 6 )
!
!     CODING FOR FACTOR 6
!     -------------------
      IIA=1
      IIB=IIA+2*IM-KLA
      IIC=IIB+2*IM
      IID=IIC+2*IM
      IIE=IIC
      IIF=IIB
      IJA=1
      IJB=IJA+JINK
      IJC=IJB+JINK
      IJD=IJC+JINK
      IJE=IJD+JINK
      IJF=IJE+JINK

      IF ( KLA /= IM ) THEN

!$acc kernels
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=(PA(IIA+II)+PA(IID+II))+(PA(IIB+II)+PA(IIC+II))
            PC(IJD+IJ)=(PA(IIA+II)-PA(IID+II))-(PA(IIB+II)-PA(IIC+II))
            PC(IJB+IJ)=((PA(IIA+II)-PA(IID+II))+0.5*(PA(IIB+II)-PA(IIC+II))) &
                -(XSIN60*(PB(IIB+II)+PB(IIC+II)))
            PC(IJF+IJ)=((PA(IIA+II)-PA(IID+II))+0.5*(PA(IIB+II)-PA(IIC+II))) &
                +(XSIN60*(PB(IIB+II)+PB(IIC+II)))
            PC(IJC+IJ)=((PA(IIA+II)+PA(IID+II))-0.5*(PA(IIB+II)+PA(IIC+II))) &
                -(XSIN60*(PB(IIB+II)-PB(IIC+II)))
            PC(IJE+IJ)=((PA(IIA+II)+PA(IID+II))-0.5*(PA(IIB+II)+PA(IIC+II))) &
                +(XSIN60*(PB(IIB+II)-PB(IIC+II)))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IIA=IIA+IINK
        IINK=2*IINK
        IIB=IIB+IINK
        IIC=IIC+IINK
        IID=IID-IINK
        IIE=IIE-IINK
        IIF=IIF-IINK
        IJBASE=IJBASE+JUMP
        JUMP=2*JUMP+JINK
        IF ( IIC /= IID ) THEN
!$acc kernels
          IJBASE0 = IJBASE
          IIA0 = IIA
          IIB0 = IIB
          IIC0 = IIC
          IID0 = IID
          IIE0 = IIE
          IIF0 = IIF
!$acc loop independent private( IKB, IKC, IKD, IKE, IKF, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, ZC4, ZS4, ZC5, ZS5, &
!$acc &                         IJBASE, IIA, IIB, IIC, IID, IIE, IIF )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            IKD=IKC+IKB
            IKE=IKD+IKB
            IKF=IKE+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            ZC3=PTRIGS(IKD+1)
            ZS3=PTRIGS(IKD+2)
            ZC4=PTRIGS(IKE+1)
            ZS4=PTRIGS(IKE+2)
            ZC5=PTRIGS(IKF+1)
            ZS5=PTRIGS(IKF+2)
            IJBASE = IJBASE0 + (IIK-KLA)/KLA * (KLA + JUMP )
            IIA = IIA0 + (IIK-KLA)/KLA * IINK
            IIB = IIB0 + (IIK-KLA)/KLA * IINK
            IIC = IIC0 + (IIK-KLA)/KLA * IINK
            IID = IID0 - (IIK-KLA)/KLA * IINK
            IIE = IIE0 - (IIK-KLA)/KLA * IINK
            IIF = IIF0 - (IIK-KLA)/KLA * IINK
!$acc loop independent
            DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA11, ZA20, ZA21, ZB11, ZB20, ZB21 )
              DO IJK=1,KLOT
                II =         IIL - 1 + (IJK - 1 ) * KINC3
                IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4

                ZA11= (PA(IIE+II)+PA(IIB+II))+(PA(IIC+II)+PA(IIF+II))
                ZA20=(PA(IIA+II)+PA(IID+II))-0.5*ZA11
                ZA21=XSIN60*((PA(IIE+II)+PA(IIB+II))-(PA(IIC+II)+PA(IIF+II)))
                ZB11= (PB(IIB+II)-PB(IIE+II))+(PB(IIC+II)-PB(IIF+II))
                ZB20=(PB(IIA+II)-PB(IID+II))-0.5*ZB11
                ZB21=XSIN60*((PB(IIB+II)-PB(IIE+II))-(PB(IIC+II)-PB(IIF+II)))

                PC(IJA+IJ)=(PA(IIA+II)+PA(IID+II))+ZA11
                PD(IJA+IJ)=(PB(IIA+II)-PB(IID+II))+ZB11
                PC(IJC+IJ)=ZC2*(ZA20-ZB21)-ZS2*(ZB20+ZA21)
                PD(IJC+IJ)=ZS2*(ZA20-ZB21)+ZC2*(ZB20+ZA21)
                PC(IJE+IJ)=ZC4*(ZA20+ZB21)-ZS4*(ZB20-ZA21)
                PD(IJE+IJ)=ZS4*(ZA20+ZB21)+ZC4*(ZB20-ZA21)

                ZA11=(PA(IIE+II)-PA(IIB+II))+(PA(IIC+II)-PA(IIF+II))
                ZB11=(PB(IIE+II)+PB(IIB+II))-(PB(IIC+II)+PB(IIF+II))
                ZA20=(PA(IIA+II)-PA(IID+II))-0.5*ZA11
                ZA21=XSIN60*((PA(IIE+II)-PA(IIB+II))-(PA(IIC+II)-PA(IIF+II)))
                ZB20=(PB(IIA+II)+PB(IID+II))+0.5*ZB11
                ZB21=XSIN60*((PB(IIE+II)+PB(IIB+II))+(PB(IIC+II)+PB(IIF+II)))

                PC(IJD+IJ)=                                                          &
                  ZC3*((PA(IIA+II)-PA(IID+II))+ZA11)-ZS3*((PB(IIA+II)+PB(IID+II))-ZB11)
                PD(IJD+IJ)=                                                          &
                  ZS3*((PA(IIA+II)-PA(IID+II))+ZA11)+ZC3*((PB(IIA+II)+PB(IID+II))-ZB11)
                PC(IJB+IJ)=ZC1*(ZA20-ZB21)-ZS1*(ZB20-ZA21)
                PD(IJB+IJ)=ZS1*(ZA20-ZB21)+ZC1*(ZB20-ZA21)
                PC(IJF+IJ)=ZC5*(ZA20+ZB21)-ZS5*(ZB20+ZA21)
                PD(IJF+IJ)=ZS5*(ZA20+ZB21)+ZC5*(ZB20+ZA21)
              END DO
            END DO
          END DO
          IJBASE = IJBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + JUMP )
          IIA = IIA0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIB = IIB0 + ((ISTOP-KLA)/KLA+1) * IINK
          IIC = IIC0 + ((ISTOP-KLA)/KLA+1) * IINK
          IID = IID0 - ((ISTOP-KLA)/KLA+1) * IINK
          IIE = IIE0 - ((ISTOP-KLA)/KLA+1) * IINK
          IIF = IIF0 - ((ISTOP-KLA)/KLA+1) * IINK
!$acc end kernels
        END IF

        IF ( IIC <= IID ) THEN
!$acc kernels
          IIBASE=0
!$acc loop independent
          DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIB+II)+(PA(IIA+II)+PA(IIC+II))
              PC(IJD+IJ)=PB(IIB+II)-(PB(IIA+II)+PB(IIC+II))
              PC(IJB+IJ)=(XSIN60*(PA(IIA+II)-PA(IIC+II)))-(0.5*(PB(IIA+II)+PB(IIC+II))+PB(IIB+ &
                II))
              PC(IJF+IJ)=-(XSIN60*(PA(IIA+II)-PA(IIC+II)))-(0.5*(PB(IIA+II)+PB(IIC+II))+PB(IIB+ &
                II))
              PC(IJC+IJ)=XSIN60*(PB(IIC+II)-PB(IIA+II))+(0.5*(PA(IIA+II)+PA(IIC+II))-PA(IIB+II))
              PC(IJE+IJ)=XSIN60*(PB(IIC+II)-PB(IIA+II))-(0.5*(PA(IIA+II)+PA(IIC+II))-PA(IIB+II))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
        ZSSIN60=2.0*XSIN60
!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=(2.0*(PA(IIA+II)+PA(IID+II)))+(2.0*(PA(IIB+II)+PA(IIC+II)))
            PC(IJD+IJ)=(2.0*(PA(IIA+II)-PA(IID+II)))-(2.0*(PA(IIB+II)-PA(IIC+II)))
            PC(IJB+IJ)=(2.0*(PA(IIA+II)-PA(IID+II))+(PA(IIB+II)-PA(IIC+II))) &
                -(ZSSIN60*(PB(IIB+II)+PB(IIC+II)))
            PC(IJF+IJ)=(2.0*(PA(IIA+II)-PA(IID+II))+(PA(IIB+II)-PA(IIC+II))) &
                +(ZSSIN60*(PB(IIB+II)+PB(IIC+II)))
            PC(IJC+IJ)=(2.0*(PA(IIA+II)+PA(IID+II))-(PA(IIB+II)+PA(IIC+II))) &
                -(ZSSIN60*(PB(IIB+II)-PB(IIC+II)))
            PC(IJE+IJ)=(2.0*(PA(IIA+II)+PA(IID+II))-(PA(IIB+II)+PA(IIC+II))) &
                +(ZSSIN60*(PB(IIB+II)-PB(IIC+II)))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF

      IBAD=0


      CASE ( 7, 8 )
!
!     CODING FOR FACTOR 8
!     -------------------
      IF ( KLA /= IM ) THEN
        IBAD = 3
      ELSE
!$acc kernels
        IIA=1
        IIB=IIA+KLA
        IIC=IIB+2*KLA
        IID=IIC+2*KLA
        IIE=IID+2*KLA
        IJA=1
        IJB=IJA+JINK
        IJC=IJB+JINK
        IJD=IJC+JINK
        IJE=IJD+JINK
        IJF=IJE+JINK
        IJG=IJF+JINK
        IJH=IJG+JINK
        ZSSIN45=SQRT(2.0)

!$acc loop independent
        DO IIL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IIL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IIL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=2.0*(((PA(IIA+II)+PA(IIE+II))+PA(IIC+II))+(PA(IIB+II)+PA(IID+II)))
            PC(IJE+IJ)=2.0*(((PA(IIA+II)+PA(IIE+II))+PA(IIC+II))-(PA(IIB+II)+PA(IID+II)))
            PC(IJC+IJ)=2.0*(((PA(IIA+II)+PA(IIE+II))-PA(IIC+II))-(PB(IIB+II)-PB(IID+II)))
            PC(IJG+IJ)=2.0*(((PA(IIA+II)+PA(IIE+II))-PA(IIC+II))+(PB(IIB+II)-PB(IID+II)))
            PC(IJB+IJ)=2.0*((PA(IIA+II)-PA(IIE+II))-PB(IIC+II))           &
                +ZSSIN45*((PA(IIB+II)-PA(IID+II))-(PB(IIB+II)+PB(IID+II)))
            PC(IJF+IJ)=2.0*((PA(IIA+II)-PA(IIE+II))-PB(IIC+II))           &
                -ZSSIN45*((PA(IIB+II)-PA(IID+II))-(PB(IIB+II)+PB(IID+II)))
            PC(IJD+IJ)=2.0*((PA(IIA+II)-PA(IIE+II))+PB(IIC+II))           &
                -ZSSIN45*((PA(IIB+II)-PA(IID+II))+(PB(IIB+II)+PB(IID+II)))
            PC(IJH+IJ)=2.0*((PA(IIA+II)-PA(IIE+II))+PB(IIC+II))           &
                +ZSSIN45*((PA(IIB+II)-PA(IID+II))+(PB(IIB+II)+PB(IID+II)))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IBAD=0
      END IF !KLA==M

      CASE DEFAULT
        CALL PRINT_MSG( NVERB_ERROR, 'GEN', 'RPASSM', 'invalid factor' )

      END SELECT
!
!     RETURN
!     ------
910 continue
      KERR=IBAD

!$acc end data


END SUBROUTINE RPASSM



SUBROUTINE QPASSM( PA, PB, PC, PD, PTRIGS, KINC3, KINC4, KLOT, KN, KFAC, KLA, KERR )
  IMPLICIT NONE

  REAL, DIMENSION(:), INTENT(IN)    :: PA
  REAL, DIMENSION(:), INTENT(IN)    :: PB
  REAL, DIMENSION(:), INTENT(INOUT) :: PC
  REAL, DIMENSION(:), INTENT(INOUT) :: PD
  REAL, DIMENSION(:), INTENT(IN)    :: PTRIGS
  INTEGER,            INTENT(IN)    :: KINC3, KINC4, KLOT, KN, KFAC, KLA
  INTEGER,            INTENT(OUT)   :: KERR
!
!     SUBROUTINE 'QPASSM' - PERFORMS ONE PASS THROUGH DATA AS PART
!     OF MULTIPLE REAL FFT (FOURIER ANALYSIS) ROUTINE
!
!     PA IS FIRST REAL INPUT VECTOR
!         EQUIVALENCE PB(1) WITH PA(KFAC*KLA+1)
!     PC IS FIRST REAL OUTPUT VECTOR
!         EQUIVALENCE PD(1) WITH PC(KLA+1)
!     PTRIGS IS PA PRECALCULATED LIST OF SINES & COSINES
!     KINC3 IS THE INCREMENT BETWEEN INPUT VECTORS PA
!     KINC4 IS THE INCREMENT BETWEEN OUTPUT VECTORS PC
!     KLOT IS THE NUMBER OF VECTORS
!     KN IS THE LENGTH OF THE VECTORS
!     KFAC IS THE CURRENT FACTOR OF KN
!     KLA = KN/(PRODUCT OF FACTORS USED SO FAR)
!     KERR IS AN ERROR INDICATOR:
!              0 - PASS COMPLETED WITHOUT ERROR
!              1 - KLOT GREATER THAN NVECLEN
!              2 - KFAC NOT CATERED FOR
!              3 - KFAC ONLY CATERED FOR IF KLA=KN/KFAC
!
!-----------------------------------------------------------------------
!

  INTEGER :: IM, IINK, JINK, IJUMP, ISTOP
  INTEGER :: IBAD, IIBASE, IJBASE, IGO
  INTEGER :: II, IJ, IIA, IIB, IJA, IJB
  INTEGER :: IJL, IJK, IIK, IKB, IIC, IJC, IKC
  INTEGER :: IID, IJD, IKD, IIE, IJE, IKE, IIF, IJF, IKF, IIG, IIH
  INTEGER :: IBASE0, IJA0, IJB0, IJC0, IJD0, IJE0, IJF0
  REAL :: ZC1, ZC2, ZC3, ZC4, ZC5, ZS1, ZS2, ZS3, ZS4, ZS5, ZZ
  REAL :: ZA0, ZA1, ZA2, ZA3, ZA4, ZA5, ZA6, ZA10, ZA11, ZA20, ZA21
  REAL :: ZB0, ZB1, ZB2, ZB3, ZB4, ZB5, ZB6, ZB10, ZB11, ZB20, ZB21
  REAL :: ZSIN45, ZZQRT5, ZZSIN36, ZZSIN45, ZZSIN60, ZZSIN72

!$acc data present( PA, PB, PC, PD ) copyin( PTRIGS )



      IM=KN/KFAC
      IINK=KLA
      JINK=KLA
      IJUMP=(KFAC-1)*IINK
      ISTOP=(KN-KFAC)/(2*KFAC)

      IBAD=1
      IF (KLOT.GT.NVECLEN) GO TO 910
      IIBASE=0
      IJBASE=0
      IGO=KFAC-1
      IF (IGO.EQ.7) IGO=6
      IBAD=2
      IF (IGO.LT.1.OR.IGO.GT.6) GO TO 910

      SELECT CASE( KFAC )
!
!     CODING FOR FACTOR 2
!     -------------------
      CASE ( 2 )
      IIA=1
      IIB=IIA+IINK
      IJA=1
      IJB=IJA+2*IM-KLA

      IF ( KLA /= IM ) THEN

!$acc kernels
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=PA(IIA+II)+PA(IIB+II)
            PC(IJB+IJ)=PA(IIA+II)-PA(IIB+II)
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IJA=IJA+JINK
        JINK=2*JINK
        IJB=IJB-JINK
        IIBASE=IIBASE+IJUMP
        IJUMP=2*IJUMP+IINK
        IF ( IJA /= IJB ) THEN
!$acc kernels
          IBASE0 = IIBASE
          IJA0 = IJA
          IJB0 = IJB
!$acc loop independent private( IKB, ZC1, ZS1, IIBASE, IJA, IJB )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            IIBASE = IBASE0 + (IIK-KLA)/KLA * (KLA + IJUMP )
            IJA = IJA0 + (IIK-KLA)/KLA * JINK
            IJB = IJB0 - (IIK-KLA)/KLA * JINK
!$acc loop independent
            DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
              DO IJK=1,KLOT
                II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
                IJ =         IJL - 1 + (IJK - 1 ) * KINC4
                PC(IJA+IJ)=PA(IIA+II)+(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))
                PC(IJB+IJ)=PA(IIA+II)-(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))
                PD(IJA+IJ)=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))+PB(IIA+II)
                PD(IJB+IJ)=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))-PB(IIA+II)
              END DO
            END DO
          END DO
          IIBASE = IBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + IJUMP )
          IJA = IJA0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJB = IJB0 - ((ISTOP-KLA)/KLA+1) * JINK
!$acc end kernels
        END IF

        IF ( IJA <= IJB ) THEN
!$acc kernels
          IJBASE=0
!$acc loop independent
          DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIA+II)
              PD(IJA+IJ)=-PA(IIB+II)
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
        ZZ=1.0/REAL(KN)
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=ZZ*(PA(IIA+II)+PA(IIB+II))
            PC(IJB+IJ)=ZZ*(PA(IIA+II)-PA(IIB+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels

      END IF

      IBAD=0



      CASE ( 3 )
!
!     CODING FOR FACTOR 3
!C     -------------------
      IIA=1
      IIB=IIA+IINK
      IIC=IIB+IINK
      IJA=1
      IJB=IJA+2*IM-KLA
      IJC=IJB

      IF ( KLA /= IM ) THEN

!$acc kernels
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=PA(IIA+II)+(PA(IIB+II)+PA(IIC+II))
            PC(IJB+IJ)=PA(IIA+II)-0.5*(PA(IIB+II)+PA(IIC+II))
            PD(IJB+IJ)=XSIN60*(PA(IIC+II)-PA(IIB+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IJA=IJA+JINK
        JINK=2*JINK
        IJB=IJB+JINK
        IJC=IJC-JINK
        IIBASE=IIBASE+IJUMP
        IJUMP=2*IJUMP+IINK
        IF ( IJA /= IJC ) THEN
!$acc kernels
          IBASE0 = IIBASE
          IJA0 = IJA
          IJB0 = IJB
          IJC0 = IJC
!$acc loop independent private( IKB, IKC, ZC1, ZS1, ZC2, ZS2, IIBASE, IJA, IJB, IJC )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            IIBASE = IBASE0 + (IIK-KLA)/KLA * (KLA + IJUMP )
            IJA = IJA0 + (IIK-KLA)/KLA * JINK
            IJB = IJB0 + (IIK-KLA)/KLA * JINK
            IJC = IJC0 - (IIK-KLA)/KLA * JINK
!$acc loop independent
            DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA1, ZA2, ZA3, ZB1, ZB2, ZB3 )
              DO IJK=1,KLOT
                II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
                IJ =         IJL - 1 + (IJK - 1 ) * KINC4
                ZA1=(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))+(ZC2*PA(IIC+II)+ZS2*PB(IIC+II))
                ZB1=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))+(ZC2*PB(IIC+II)-ZS2*PA(IIC+II))
                ZA2=PA(IIA+II)-0.5*ZA1
                ZB2=PB(IIA+II)-0.5*ZB1
                ZA3=XSIN60*((ZC1*PA(IIB+II)+ZS1*PB(IIB+II))-(ZC2*PA(IIC+II)+ZS2*PB(IIC+II)))
                ZB3=XSIN60*((ZC1*PB(IIB+II)-ZS1*PA(IIB+II))-(ZC2*PB(IIC+II)-ZS2*PA(IIC+II)))
                PC(IJA+IJ)=PA(IIA+II)+ZA1
                PD(IJA+IJ)=PB(IIA+II)+ZB1
                PC(IJB+IJ)=ZA2+ZB3
                PD(IJB+IJ)=ZB2-ZA3
                PC(IJC+IJ)=ZA2-ZB3
                PD(IJC+IJ)=-(ZB2+ZA3)
              END DO
            END DO
          END DO
          IIBASE = IBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + IJUMP )
          IJA = IJA0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJB = IJB0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJC = IJC0 - ((ISTOP-KLA)/KLA+1) * JINK
!$acc end kernels
        END IF

        IF ( IJA <= IJC ) THEN
!$acc kernels
          IJBASE=0
!$acc loop independent
          DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIA+II)+0.5*(PA(IIB+II)-PA(IIC+II))
              PD(IJA+IJ)=-XSIN60*(PA(IIB+II)+PA(IIC+II))
              PC(IJB+IJ)=PA(IIA+II)-(PA(IIB+II)-PA(IIC+II))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
        ZZ=1.0/REAL(KN)
        ZZSIN60=ZZ*XSIN60
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=ZZ*(PA(IIA+II)+(PA(IIB+II)+PA(IIC+II)))
            PC(IJB+IJ)=ZZ*(PA(IIA+II)-0.5*(PA(IIB+II)+PA(IIC+II)))
            PD(IJB+IJ)=ZZSIN60*(PA(IIC+II)-PA(IIB+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF

      IBAD=0



      CASE ( 4 )
!
!     CODING FOR FACTOR 4
!     -------------------
      IIA=1
      IIB=IIA+IINK
      IIC=IIB+IINK
      IID=IIC+IINK
      IJA=1
      IJB=IJA+2*IM-KLA
      IJC=IJB+2*IM
      IJD=IJB

      IF ( KLA /= IM ) THEN
!$acc kernels
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=(PA(IIA+II)+PA(IIC+II))+(PA(IIB+II)+PA(IID+II))
            PC(IJC+IJ)=(PA(IIA+II)+PA(IIC+II))-(PA(IIB+II)+PA(IID+II))
            PC(IJB+IJ)=PA(IIA+II)-PA(IIC+II)
            PD(IJB+IJ)=PA(IID+II)-PA(IIB+II)
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IJA=IJA+JINK
        JINK=2*JINK
        IJB=IJB+JINK
        IJC=IJC-JINK
        IJD=IJD-JINK
        IIBASE=IIBASE+IJUMP
        IJUMP=2*IJUMP+IINK
        IF ( IJB /= IJC ) THEN
!$acc kernels
          IBASE0 = IIBASE
          IJA0 = IJA
          IJB0 = IJB
          IJC0 = IJC
          IJD0 = IJD
!$acc loop independent private( IKB, IKC, IKD, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, IIBASE, IJA, IJB, IJC, IJD )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            IKD=IKC+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            ZC3=PTRIGS(IKD+1)
            ZS3=PTRIGS(IKD+2)
            IIBASE = IBASE0 + (IIK-KLA)/KLA * (KLA + IJUMP )
            IJA = IJA0 + (IIK-KLA)/KLA * JINK
            IJB = IJB0 + (IIK-KLA)/KLA * JINK
            IJC = IJC0 - (IIK-KLA)/KLA * JINK
            IJD = IJD0 - (IIK-KLA)/KLA * JINK
!$acc loop independent
            DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA0, ZA1, ZA2, ZA3, ZB0, ZB1, ZB2, ZB3 )
              DO IJK=1,KLOT
                II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
                IJ =         IJL - 1 + (IJK - 1 ) * KINC4
                ZA0=PA(IIA+II)+(ZC2*PA(IIC+II)+ZS2*PB(IIC+II))
                ZA2=PA(IIA+II)-(ZC2*PA(IIC+II)+ZS2*PB(IIC+II))
                ZA1=(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))+(ZC3*PA(IID+II)+ZS3*PB(IID+II))
                ZA3=(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))-(ZC3*PA(IID+II)+ZS3*PB(IID+II))
                ZB0=PB(IIA+II)+(ZC2*PB(IIC+II)-ZS2*PA(IIC+II))
                ZB2=PB(IIA+II)-(ZC2*PB(IIC+II)-ZS2*PA(IIC+II))
                ZB1=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))+(ZC3*PB(IID+II)-ZS3*PA(IID+II))
                ZB3=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))-(ZC3*PB(IID+II)-ZS3*PA(IID+II))
                PC(IJA+IJ)=ZA0+ZA1
                PC(IJC+IJ)=ZA0-ZA1
                PD(IJA+IJ)=ZB0+ZB1
                PD(IJC+IJ)=ZB1-ZB0
                PC(IJB+IJ)=ZA2+ZB3
                PC(IJD+IJ)=ZA2-ZB3
                PD(IJB+IJ)=ZB2-ZA3
                PD(IJD+IJ)=-(ZB2+ZA3)
              END DO
            END DO
          END DO
          IIBASE = IBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + IJUMP )
          IJA = IJA0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJB = IJB0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJC = IJC0 - ((ISTOP-KLA)/KLA+1) * JINK
          IJD = IJD0 - ((ISTOP-KLA)/KLA+1) * JINK
!$acc end kernels
        END IF

        IF ( IJB <= IJC ) THEN
!$acc kernels
          ZSIN45=SQRT(0.5)
          IJBASE=0
!$acc loop independent
          DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=PA(IIA+II)+ZSIN45*(PA(IIB+II)-PA(IID+II))
              PC(IJB+IJ)=PA(IIA+II)-ZSIN45*(PA(IIB+II)-PA(IID+II))
              PD(IJA+IJ)=-PA(IIC+II)-ZSIN45*(PA(IIB+II)+PA(IID+II))
              PD(IJB+IJ)=PA(IIC+II)-ZSIN45*(PA(IIB+II)+PA(IID+II))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE ! KLA == M
!$acc kernels
        ZZ=1.0/REAL(KN)
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=ZZ*((PA(IIA+II)+PA(IIC+II))+(PA(IIB+II)+PA(IID+II)))
            PC(IJC+IJ)=ZZ*((PA(IIA+II)+PA(IIC+II))-(PA(IIB+II)+PA(IID+II)))
            PC(IJB+IJ)=ZZ*(PA(IIA+II)-PA(IIC+II))
            PD(IJB+IJ)=ZZ*(PA(IID+II)-PA(IIB+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF
      IBAD=0



      CASE ( 5 )
!
!     CODING FOR FACTOR 5
!     -------------------
      IIA=1
      IIB=IIA+IINK
      IIC=IIB+IINK
      IID=IIC+IINK
      IIE=IID+IINK
      IJA=1
      IJB=IJA+2*IM-KLA
      IJC=IJB+2*IM
      IJD=IJC
      IJE=IJB

      IF ( KLA /= IM ) THEN
!$acc kernels
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            ZA1=PA(IIB+II)+PA(IIE+II)
            ZA3=PA(IIB+II)-PA(IIE+II)
            ZA2=PA(IIC+II)+PA(IID+II)
            ZA4=PA(IIC+II)-PA(IID+II)
            ZA5=PA(IIA+II)-0.25*(ZA1+ZA2)
            ZA6=XQRT5*(ZA1-ZA2)
            PC(IJA+IJ)=PA(IIA+II)+(ZA1+ZA2)
            PC(IJB+IJ)=ZA5+ZA6
            PC(IJC+IJ)=ZA5-ZA6
            PD(IJB+IJ)=-XSIN72*ZA3-XSIN36*ZA4
            PD(IJC+IJ)=-XSIN36*ZA3+XSIN72*ZA4
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IJA=IJA+JINK
        JINK=2*JINK
        IJB=IJB+JINK
        IJC=IJC+JINK
        IJD=IJD-JINK
        IJE=IJE-JINK
        IIBASE=IIBASE+IJUMP
        IJUMP=2*IJUMP+IINK
        IF ( IJB /= IJD ) THEN
!$acc kernels
          IBASE0 = IIBASE
          IJA0 = IJA
          IJB0 = IJB
          IJC0 = IJC
          IJD0 = IJD
          IJE0 = IJE
!$acc loop independent private( IKB, IKC, IKD, IKE, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, ZC4, ZS4, IIBASE, IJA, IJB, IJC, IJD, IJE )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            IKD=IKC+IKB
            IKE=IKD+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            ZC3=PTRIGS(IKD+1)
            ZS3=PTRIGS(IKD+2)
            ZC4=PTRIGS(IKE+1)
            ZS4=PTRIGS(IKE+2)
            IIBASE = IBASE0 + (IIK-KLA)/KLA * (KLA + IJUMP )
            IJA = IJA0 + (IIK-KLA)/KLA * JINK
            IJB = IJB0 + (IIK-KLA)/KLA * JINK
            IJC = IJC0 + (IIK-KLA)/KLA * JINK
            IJD = IJD0 - (IIK-KLA)/KLA * JINK
            IJE = IJE0 - (IIK-KLA)/KLA * JINK
!$acc loop independent
            DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA1, ZA2, ZA3, ZA4, ZA5, ZA6, ZA10, ZA11, ZA20, ZA21, &
!$acc &                         ZB1, ZB2, ZB3, ZB4, ZB5, ZB6, ZB10, ZB11, ZB20, ZB21 )
              DO IJK=1,KLOT
                II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
                IJ =         IJL - 1 + (IJK - 1 ) * KINC4
                ZA1=(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))+(ZC4*PA(IIE+II)+ZS4*PB(IIE+II))
                ZA3=(ZC1*PA(IIB+II)+ZS1*PB(IIB+II))-(ZC4*PA(IIE+II)+ZS4*PB(IIE+II))
                ZA2=(ZC2*PA(IIC+II)+ZS2*PB(IIC+II))+(ZC3*PA(IID+II)+ZS3*PB(IID+II))
                ZA4=(ZC2*PA(IIC+II)+ZS2*PB(IIC+II))-(ZC3*PA(IID+II)+ZS3*PB(IID+II))
                ZB1=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))+(ZC4*PB(IIE+II)-ZS4*PA(IIE+II))
                ZB3=(ZC1*PB(IIB+II)-ZS1*PA(IIB+II))-(ZC4*PB(IIE+II)-ZS4*PA(IIE+II))
                ZB2=(ZC2*PB(IIC+II)-ZS2*PA(IIC+II))+(ZC3*PB(IID+II)-ZS3*PA(IID+II))
                ZB4=(ZC2*PB(IIC+II)-ZS2*PA(IIC+II))-(ZC3*PB(IID+II)-ZS3*PA(IID+II))
                ZA5=PA(IIA+II)-0.25*(ZA1+ZA2)
                ZA6=XQRT5*(ZA1-ZA2)
                ZB5=PB(IIA+II)-0.25*(ZB1+ZB2)
                ZB6=XQRT5*(ZB1-ZB2)
                ZA10=ZA5+ZA6
                ZA20=ZA5-ZA6
                ZB10=ZB5+ZB6
                ZB20=ZB5-ZB6
                ZA11=XSIN72*ZB3+XSIN36*ZB4
                ZA21=XSIN36*ZB3-XSIN72*ZB4
                ZB11=XSIN72*ZA3+XSIN36*ZA4
                ZB21=XSIN36*ZA3-XSIN72*ZA4
                PC(IJA+IJ)=PA(IIA+II)+(ZA1+ZA2)
                PC(IJB+IJ)=ZA10+ZA11
                PC(IJE+IJ)=ZA10-ZA11
                PC(IJC+IJ)=ZA20+ZA21
                PC(IJD+IJ)=ZA20-ZA21
                PD(IJA+IJ)=PB(IIA+II)+(ZB1+ZB2)
                PD(IJB+IJ)=ZB10-ZB11
                PD(IJE+IJ)=-(ZB10+ZB11)
                PD(IJC+IJ)=ZB20-ZB21
                PD(IJD+IJ)=-(ZB20+ZB21)
              END DO
            END DO
          END DO
          IIBASE = IBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + IJUMP )
          IJA = IJA0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJB = IJB0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJC = IJC0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJD = IJD0 - ((ISTOP-KLA)/KLA+1) * JINK
          IJE = IJE0 - ((ISTOP-KLA)/KLA+1) * JINK
!$acc end kernels
        END IF

        IF ( IJB <= IJD ) THEN
!$acc kernels
          IJBASE=0
!$acc loop independent
          DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA1, ZA2, ZA3, ZA4, ZA5, ZA6 )
            DO IJK=1,KLOT
              II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
              ZA1=PA(IIB+II)+PA(IIE+II)
              ZA3=PA(IIB+II)-PA(IIE+II)
              ZA2=PA(IIC+II)+PA(IID+II)
              ZA4=PA(IIC+II)-PA(IID+II)
              ZA5=PA(IIA+II)+0.25*(ZA3-ZA4)
              ZA6=XQRT5*(ZA3+ZA4)
              PC(IJA+IJ)=ZA5+ZA6
              PC(IJB+IJ)=ZA5-ZA6
              PC(IJC+IJ)=PA(IIA+II)-(ZA3-ZA4)
              PD(IJA+IJ)=-XSIN36*ZA1-XSIN72*ZA2
              PD(IJB+IJ)=-XSIN72*ZA1+XSIN36*ZA2
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE
!$acc kernels
        ZZ=1.0/REAL(KN)
        ZZQRT5=ZZ*XQRT5
        ZZSIN36=ZZ*XSIN36
        ZZSIN72=ZZ*XSIN72
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA1, ZA2, ZA3, ZA4, ZA5, ZA6 )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            ZA1=PA(IIB+II)+PA(IIE+II)
            ZA3=PA(IIB+II)-PA(IIE+II)
            ZA2=PA(IIC+II)+PA(IID+II)
            ZA4=PA(IIC+II)-PA(IID+II)
            ZA5=ZZ*(PA(IIA+II)-0.25*(ZA1+ZA2))
            ZA6=ZZQRT5*(ZA1-ZA2)
            PC(IJA+IJ)=ZZ*(PA(IIA+II)+(ZA1+ZA2))
            PC(IJB+IJ)=ZA5+ZA6
            PC(IJC+IJ)=ZA5-ZA6
            PD(IJB+IJ)=-ZZSIN72*ZA3-ZZSIN36*ZA4
            PD(IJC+IJ)=-ZZSIN36*ZA3+ZZSIN72*ZA4
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF
      IBAD=0



      CASE ( 6 )
!
!     CODING FOR FACTOR 6
!     -------------------
      IIA=1
      IIB=IIA+IINK
      IIC=IIB+IINK
      IID=IIC+IINK
      IIE=IID+IINK
      IIF=IIE+IINK
      IJA=1
      IJB=IJA+2*IM-KLA
      IJC=IJB+2*IM
      IJD=IJC+2*IM
      IJE=IJC
      IJF=IJB

      IF ( KLA /= IM ) THEN

!$acc kernels
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA11 )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            ZA11=(PA(IIC+II)+PA(IIF+II))+(PA(IIB+II)+PA(IIE+II))
            PC(IJA+IJ)=(PA(IIA+II)+PA(IID+II))+ZA11
            PC(IJC+IJ)=(PA(IIA+II)+PA(IID+II)-0.5*ZA11)
            PD(IJC+IJ)=XSIN60*((PA(IIC+II)+PA(IIF+II))-(PA(IIB+II)+PA(IIE+II)))
            ZA11=(PA(IIC+II)-PA(IIF+II))+(PA(IIE+II)-PA(IIB+II))
            PC(IJB+IJ)=(PA(IIA+II)-PA(IID+II))-0.5*ZA11
            PD(IJB+IJ)=XSIN60*((PA(IIE+II)-PA(IIB+II))-(PA(IIC+II)-PA(IIF+II)))
            PC(IJD+IJ)=(PA(IIA+II)-PA(IID+II))+ZA11
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IJA=IJA+JINK
        JINK=2*JINK
        IJB=IJB+JINK
        IJC=IJC+JINK
        IJD=IJD-JINK
        IJE=IJE-JINK
        IJF=IJF-JINK
        IIBASE=IIBASE+IJUMP
        IJUMP=2*IJUMP+IINK
        IF (IJC /= IJD ) THEN
!$acc kernels
          IBASE0 = IIBASE
          IJA0 = IJA
          IJB0 = IJB
          IJC0 = IJC
          IJD0 = IJD
          IJE0 = IJE
          IJF0 = IJF
!$acc loop independent private( IKB, IKC, IKD, IKE, IKF, ZC1, ZS1, ZC2, ZS2, ZC3, ZS3, ZC4, ZS4, ZC5, ZS5, &
!$acc &                         IIBASE, IJA, IJB, IJC, IJD, IJE, IJF )
          DO IIK=KLA,ISTOP,KLA
            IKB=IIK+IIK
            IKC=IKB+IKB
            IKD=IKC+IKB
            IKE=IKD+IKB
            IKF=IKE+IKB
            ZC1=PTRIGS(IKB+1)
            ZS1=PTRIGS(IKB+2)
            ZC2=PTRIGS(IKC+1)
            ZS2=PTRIGS(IKC+2)
            ZC3=PTRIGS(IKD+1)
            ZS3=PTRIGS(IKD+2)
            ZC4=PTRIGS(IKE+1)
            ZS4=PTRIGS(IKE+2)
            ZC5=PTRIGS(IKF+1)
            ZS5=PTRIGS(IKF+2)
            IIBASE = IBASE0 + (IIK-KLA)/KLA * (KLA + IJUMP )
            IJA = IJA0 + (IIK-KLA)/KLA * JINK
            IJB = IJB0 + (IIK-KLA)/KLA * JINK
            IJC = IJC0 + (IIK-KLA)/KLA * JINK
            IJD = IJD0 - (IIK-KLA)/KLA * JINK
            IJE = IJE0 - (IIK-KLA)/KLA * JINK
            IJF = IJF0 - (IIK-KLA)/KLA * JINK
!$acc loop independent
            DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA1, ZA2, ZA3, ZA4, ZA5, ZA11, ZA20, ZA21, ZB1, ZB2, ZB3, ZB4, ZB5, ZB11, ZB20, ZB21 )
              DO IJK=1,KLOT
                II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
                IJ =         IJL - 1 + (IJK - 1 ) * KINC4
                ZA1=ZC1*PA(IIB+II)+ZS1*PB(IIB+II)
                ZB1=ZC1*PB(IIB+II)-ZS1*PA(IIB+II)
                ZA2=ZC2*PA(IIC+II)+ZS2*PB(IIC+II)
                ZB2=ZC2*PB(IIC+II)-ZS2*PA(IIC+II)
                ZA3=ZC3*PA(IID+II)+ZS3*PB(IID+II)
                ZB3=ZC3*PB(IID+II)-ZS3*PA(IID+II)
                ZA4=ZC4*PA(IIE+II)+ZS4*PB(IIE+II)
                ZB4=ZC4*PB(IIE+II)-ZS4*PA(IIE+II)
                ZA5=ZC5*PA(IIF+II)+ZS5*PB(IIF+II)
                ZB5=ZC5*PB(IIF+II)-ZS5*PA(IIF+II)
                ZA11=(ZA2+ZA5)+(ZA1+ZA4)
                ZA20=(PA(IIA+II)+ZA3)-0.5*ZA11
                ZA21=XSIN60*((ZA2+ZA5)-(ZA1+ZA4))
                ZB11=(ZB2+ZB5)+(ZB1+ZB4)
                ZB20=(PB(IIA+II)+ZB3)-0.5*ZB11
                ZB21=XSIN60*((ZB2+ZB5)-(ZB1+ZB4))
                PC(IJA+IJ)=(PA(IIA+II)+ZA3)+ZA11
                PD(IJA+IJ)=(PB(IIA+II)+ZB3)+ZB11
                PC(IJC+IJ)=ZA20-ZB21
                PD(IJC+IJ)=ZA21+ZB20
                PC(IJE+IJ)=ZA20+ZB21
                PD(IJE+IJ)=ZA21-ZB20
                ZA11=(ZA2-ZA5)+(ZA4-ZA1)
                ZA20=(PA(IIA+II)-ZA3)-0.5*ZA11
                ZA21=XSIN60*((ZA4-ZA1)-(ZA2-ZA5))
                ZB11=(ZB5-ZB2)-(ZB4-ZB1)
                ZB20=(ZB3-PB(IIA+II))-0.5*ZB11
                ZB21=XSIN60*((ZB5-ZB2)+(ZB4-ZB1))
                PC(IJB+IJ)=ZA20-ZB21
                PD(IJB+IJ)=ZA21-ZB20
                PC(IJD+IJ)=ZA11+(PA(IIA+II)-ZA3)
                PD(IJD+IJ)=ZB11+(ZB3-PB(IIA+II))
                PC(IJF+IJ)=ZA20+ZB21
                PD(IJF+IJ)=ZA21+ZB20
              END DO
            END DO
          END DO
          IIBASE = IBASE0 + ((ISTOP-KLA)/KLA+1) * ( KLA + IJUMP )
          IJA = IJA0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJB = IJB0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJC = IJC0 + ((ISTOP-KLA)/KLA+1) * JINK
          IJD = IJD0 - ((ISTOP-KLA)/KLA+1) * JINK
          IJE = IJE0 - ((ISTOP-KLA)/KLA+1) * JINK
          IJF = IJF0 - ((ISTOP-KLA)/KLA+1) * JINK
!$acc end kernels
        END IF

        IF ( IJC <= IJD ) THEN
!$acc kernels
          IJBASE=0
!$acc loop independent
          DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
            DO IJK=1,KLOT
              II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
              IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
              PC(IJA+IJ)=(PA(IIA+II)+0.5*(PA(IIC+II)-PA(IIE+II)))+ XSIN60*(PA(IIB+II)-PA(IIF+II))
              PD(IJA+IJ)=-(PA(IID+II)+0.5*(PA(IIB+II)+PA(IIF+II)))-XSIN60*(PA(IIC+II)+PA(IIE+II))
              PC(IJB+IJ)=PA(IIA+II)-(PA(IIC+II)-PA(IIE+II))
              PD(IJB+IJ)=PA(IID+II)-(PA(IIB+II)+PA(IIF+II))
              PC(IJC+IJ)=(PA(IIA+II)+0.5*(PA(IIC+II)-PA(IIE+II)))-XSIN60*(PA(IIB+II)-PA(IIF+II))
              PD(IJC+IJ)=-(PA(IID+II)+0.5*(PA(IIB+II)+PA(IIF+II)))+XSIN60*(PA(IIC+II)+PA(IIE+II))
            END DO
          END DO
          IIBASE = IIBASE + KLA
          IJBASE = IJBASE + KLA
!$acc end kernels
        END IF

      ELSE

!$acc kernels
        ZZ=1.0/REAL(KN)
        ZZSIN60=ZZ*XSIN60
!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ, ZA11 )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            ZA11=(PA(IIC+II)+PA(IIF+II))+(PA(IIB+II)+PA(IIE+II))
            PC(IJA+IJ)=ZZ*((PA(IIA+II)+PA(IID+II))+ZA11)
            PC(IJC+IJ)=ZZ*((PA(IIA+II)+PA(IID+II))-0.5*ZA11)
            PD(IJC+IJ)=ZZSIN60*((PA(IIC+II)+PA(IIF+II))-(PA(IIB+II)+PA(IIE+II)))
            ZA11=(PA(IIC+II)-PA(IIF+II))+(PA(IIE+II)-PA(IIB+II))
            PC(IJB+IJ)=ZZ*((PA(IIA+II)-PA(IID+II))-0.5*ZA11)
            PD(IJB+IJ)=ZZSIN60*((PA(IIE+II)-PA(IIB+II))-(PA(IIC+II)-PA(IIF+II)))
            PC(IJD+IJ)=ZZ*((PA(IIA+II)-PA(IID+II))+ZA11)
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
      END IF

      IBAD=0



      CASE ( 7, 8 )
!
!     CODING FOR FACTOR 8
!     -------------------
      IF ( KLA /= IM ) THEN
        IBAD = 3
      ELSE
!$acc kernels
        IIA=1
        IIB=IIA+IINK
        IIC=IIB+IINK
        IID=IIC+IINK
        IIE=IID+IINK
        IIF=IIE+IINK
        IIG=IIF+IINK
        IIH=IIG+IINK
        IJA=1
        IJB=IJA+KLA
        IJC=IJB+2*IM
        IJD=IJC+2*IM
        IJE=IJD+2*IM
        ZZ=1.0/REAL(KN)
        ZZSIN45=ZZ*SQRT(0.5)

!$acc loop independent
        DO IJL=1,KLA
!CDIR$ IVDEP
!!CDIR NODEP
!*VOCL LOOP,NOVREC
!$acc loop independent private( II, IJ )
          DO IJK=1,KLOT
            II = IIBASE + IJL - 1 + (IJK - 1 ) * KINC3
            IJ = IJBASE + IJL - 1 + (IJK - 1 ) * KINC4
            PC(IJA+IJ)=ZZ*(((PA(IIA+II)+PA(IIE+II))+(PA(IIC+II)+PA(IIG+II)))+ &
                ((PA(IID+II)+PA(IIH+II))+(PA(IIB+II)+PA(IIF+II))))
            PC(IJE+IJ)=ZZ*(((PA(IIA+II)+PA(IIE+II))+(PA(IIC+II)+PA(IIG+II)))- &
                ((PA(IID+II)+PA(IIH+II))+(PA(IIB+II)+PA(IIF+II))))
            PC(IJC+IJ)=ZZ*((PA(IIA+II)+PA(IIE+II))-(PA(IIC+II)+PA(IIG+II)))
            PD(IJC+IJ)=ZZ*((PA(IID+II)+PA(IIH+II))-(PA(IIB+II)+PA(IIF+II)))
            PC(IJB+IJ)=ZZ*(PA(IIA+II)-PA(IIE+II))                       &
                +ZZSIN45*((PA(IIH+II)-PA(IID+II))-(PA(IIF+II)-PA(IIB+II)))
            PC(IJD+IJ)=ZZ*(PA(IIA+II)-PA(IIE+II))                       &
                -ZZSIN45*((PA(IIH+II)-PA(IID+II))-(PA(IIF+II)-PA(IIB+II)))
            PD(IJB+IJ)=ZZSIN45*((PA(IIH+II)-PA(IID+II))+(PA(IIF+II)-PA(IIB+II))) &
                +ZZ*(PA(IIG+II)-PA(IIC+II))
            PD(IJD+IJ)=ZZSIN45*((PA(IIH+II)-PA(IID+II))+(PA(IIF+II)-PA(IIB+II))) &
                -ZZ*(PA(IIG+II)-PA(IIC+II))
          END DO
        END DO
        IIBASE = IIBASE + KLA
        IJBASE = IJBASE + KLA
!$acc end kernels
        IBAD=0
      END IF



      CASE DEFAULT
        CALL PRINT_MSG( NVERB_ERROR, 'GEN', 'QPASSM', 'invalid factor' )

      END SELECT
!
!     RETURN
!     ------
  910 CONTINUE
      KERR=IBAD

!$acc end data

END SUBROUTINE QPASSM

END MODULE MODE_FFT
