!MNH_LIC Copyright 1994-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ####################
      MODULE MODI_FLAT_INV
!     ####################
!
INTERFACE
!
      SUBROUTINE FLAT_INV(HLBCX,HLBCY,PDXHATM,PDYHATM,PRHOM,PAF,PBF,PCF, &
      PTRIGSX,PTRIGSY,KIFAXX,KIFAXY,PY,PF_1_Y)
!
!  
IMPLICIT NONE
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX    ! x-direction LBC type 
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY    ! y-direction LBC type 
!
REAL, INTENT(IN) :: PDXHATM                     ! mean grid increment in the x
                                                ! direction
REAL, INTENT(IN) :: PDYHATM                     ! mean grid increment in the y
                                                ! direction
!
REAL, DIMENSION (:), INTENT(IN) :: PRHOM         !  mean of XRHODJ on the plane x y 
                                                 !  localized at a mass level
!
REAL, DIMENSION(:), INTENT(IN)     :: PAF,PCF    ! vectors giving the nonvanishing
REAL, DIMENSION(:,:,:), INTENT(IN) :: PBF        ! elements of the tri-diag. 
                                                 ! matrix in the pressure eq.
!
                                                 ! arrays of sin or cos values
                                                 ! for the FFT :
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSX        ! - along x
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSY        ! - along y
!
                                                 ! decomposition in prime 
                                                 ! numbers for the FFT:
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXX      ! - along x
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXY      ! - along y
!
REAL, DIMENSION(:,:,:), INTENT(IN) :: PY         ! RHS of the equation
!                                                         
REAL, DIMENSION(:,:,:), INTENT(OUT):: PF_1_Y     ! solution of the equation
!
END SUBROUTINE FLAT_INV
!
END INTERFACE
!
END MODULE MODI_FLAT_INV
!     ######################################################################
      SUBROUTINE FLAT_INV(HLBCX,HLBCY,PDXHATM,PDYHATM,PRHOM,PAF,PBF,PCF,   &
      PTRIGSX,PTRIGSY,KIFAXX,KIFAXY,PY,PF_1_Y)
!     ######################################################################
!
!!****  *FLAT_INV * - Invert the flat quasi-laplacian operator
!!
!!    PURPOSE
!!    -------
!       This routine solves the following equation:
!                    F ( F_1_Y ) =  Y 
!     where F represents the quasi-laplacian without orography. The solution is 
!     F_1_Y.
!
!!**  METHOD
!!    ------
!!      The horizontal part of F is inverted with a FFT transform. For each 
!!    horizontal direction, the FFT form depends on the lateral boundary 
!!    conditions :
!!    - CRAY intrinsic function RFFTMLT in the cyclic case 
!!    - fast cosine transform called FFT55 for all other boundary condtions.
!!    Then, in the wavenumber space, we invert for each 
!!    horizontal mode i,j a tridiagonal matrix by a classical double sweep 
!!    method.  The singular mean mode (i,j)=(0,0) corresponds to the 
!!    undetermination of the pressure to within a constant and is treated apart.
!!    To fix this degree of freedom, we set the horizontal mean value of the 
!!    pressure perturbation to 0 at the upper level of the model. 
!!       
!!    EXTERNAL
!!    --------
!!      Subroutine FFT55 : aplly multiple fast real staggered (shifted) 
!!    cosine transform
!!      Subroutine RFFTMLT : apply real-to-complex or complex-to-real Fast 
!!    Fourier Transform (FFT) on multiple input vectors.
!!      Subroutine FFT991  : equivalent to RFFTMLT
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------ 
!!      Module MODD_PARAMETERS: declaration of parameter variables
!!        JPHEXT, JPVEXT: define the number of marginal points out of the 
!!        physical domain along horizontal and vertical directions respectively
!!      Module MODD_CONF: model configurations
!!        L2D: logical for 2D model version
!!
!!    REFERENCE
!!    ---------
!!      Book2 of documentation (subroutine FLAT_INV)     
!!
!!    AUTHOR
!!    ------
!!	P. Hereil and J. Stein       * Meteo France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original    20/07/94 
!!      Revision  Jabouille (juillet 96) replace the CRAY intrinsic function
!!                             RFFTMLT by the arpege routine FFT991
!!                  17/07/97  ( J. Stein and V. Masson) initialize the corner
!!                             verticals
!!                  17/07/97  ( J. Stein and V. Masson) initialize the corner
!!                             verticals
!!      Revision  Jabouille (septembre 97) suppress the particular case for
!!                                         tridiagonal inversion
!!                Stein     ( January 98 ) faster computation for the unused 
!!                                points under the ground and out of the domain  
!!      Modification   Lugato, Guivarch (June 1998) Parallelisation
!!                     Escobar, Stein   (July 2000) optimisation
!  P. Wautelet 22/11/2022: modernization of FFts (fixed-format F77 -> free-format F95)
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_ARGSLIST_ll, ONLY : LIST_ll
USE MODD_CONF
USE MODD_PARAMETERS
!
USE MODE_FFT,         ONLY: FFT991
USE MODE_FFT55,       ONLY: FFT55
USE MODE_ll
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,   ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
use mode_mppdb
!
!
IMPLICIT NONE
!
!*       0.1   declarations of arguments
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX    ! x-direction LBC type 
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY    ! y-direction LBC type 
!
REAL, INTENT(IN) :: PDXHATM                     ! mean grid increment in the x
                                                ! direction
REAL, INTENT(IN) :: PDYHATM                     ! mean grid increment in the y
                                                ! direction
!
REAL, DIMENSION (:), INTENT(IN) :: PRHOM         !  mean of XRHODJ on the plane x y 
                                                 !  localized at a mass level
!
REAL, DIMENSION(:), INTENT(IN)     :: PAF,PCF    ! vectors giving the nonvanishing
REAL, DIMENSION(:,:,:), INTENT(IN) :: PBF        ! elements of the tri-diag. 
                                                 ! matrix in the pressure eq.
!
                                                 ! arrays of sin or cos values
                                                 ! for the FFT :
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSX        ! - along x
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSY        ! - along y
!
                                                 ! decomposition in prime 
                                                 ! numbers for the FFT:
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXX      ! - along x
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXY      ! - along y
!
REAL, DIMENSION(:,:,:), INTENT(IN) :: PY         ! RHS of the equation
!                                                         
REAL, DIMENSION(:,:,:), INTENT(OUT):: PF_1_Y     ! solution of the equation
!
!*       0.2   declaration of local variables
!
#ifndef MNH_OPENACC
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZY  ! work array to store the RHS of the equation
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZAF ! work array to expand PAF
#else
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZY ! work array to store  the RHS of the equation
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZAF ! work array to expand PAF
#endif
INTEGER :: IIB          ! indice I for the first inner mass point along x
INTEGER :: IIE          ! indice I for the last inner mass point along x
INTEGER :: IIMAX        ! number of inner mass points along the x direction
INTEGER :: IJB          ! indice J for the first inner mass point along y
INTEGER :: IJE          ! indice J for the last inner mass point along y
INTEGER :: IJMAX        ! number of inner mass points along the y direction
INTEGER :: IKB          ! indice K for the first inner mass point along z
INTEGER :: IKE          ! indice K for the last inner mass point along z
INTEGER :: IKU          ! size of the arrays along z
INTEGER :: IKMAX        ! number of inner mass points along the z direction
!
REAL :: ZDXM2,ZDYM2     ! respectively equal to PDXHATM*PDXHATM 
                        ! and PDYHATM*PDYHATM                         
INTEGER :: JI,JJ,JK     ! loop indexes along x, y, z respectively
!
!
INTEGER :: IIE_INT,IJE_INT   ! highest indice I  and J values for the x y modes.
                             ! They  depend on the l.b.c. !
!
INTEGER :: ILOTX,ILOTY ! number of data vectors along x, y resp. computed 
                       ! in parallel during the FFT process 
!
INTEGER :: INC2X,INC2Y ! increment between the start of one data vector and 
                       ! the next for the FFT along x,y resp.
!
!
#ifndef MNH_OPENACC
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZWORKX ! work array used by
! the FFT. It has been enlarged in order to be sufficient for 2D and 3D cases
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZWORKY ! work array used by
! the FFT. It has been enlarged in order to be sufficient for 2D and 3D cases
!
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZGAM
                                                        ! intermediate arrays
REAL, DIMENSION(:,:), ALLOCATABLE   :: ZBETX            ! for the tridiag.
                                                        ! matrix inversion
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZBAND_X  ! array in X slices distribution
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZBAND_Y  ! array in Y slices distribution
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZBAND_YR ! array in Y slices distribution
!
INTEGER :: IINFO_ll           ! return code of parallel routine
!
INTEGER :: IIX,IJX,IIY,IJY ! dimensions of the extended x or y slices subdomain
!
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZBAND_YT  ! array in Y slices distribution transpose
REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: ZBAND_YRT ! array in Y slices distribution transpose
#else
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZWORKX ! work array used by
! the FFT. It has been enlarged in order to be sufficient for 2D and 3D cases
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZWORKY ! work array used by
! the FFT. It has been enlarged in order to be sufficient for 2D and 3D cases
!
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZGAM
! intermediate arrays
REAL, DIMENSION(:,:), POINTER, CONTIGUOUS   :: ZBETX            ! for the tridiag.
! matrix inversion
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZBAND_X  ! array in X slices distribution
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZBAND_Y  ! array in Y slices distribution
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZBAND_YR ! array in Y slices distribution
!
INTEGER :: IINFO_ll           ! return code of parallel routine
!
INTEGER :: IIX,IJX,IIY,IJY ! dimensions of the extended x or y slices subdomain
!
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZBAND_YT  ! array in Y slices distribution transpose
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZBAND_YRT ! array in Y slices distribution transpose
#endif
REAL, DIMENSION(:), POINTER, CONTIGUOUS :: ZBAND1D
REAL, DIMENSION(:), POINTER, CONTIGUOUS :: ZWORK1D
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK( PRHOM, "FLAT_INV beg:PRHOM" )
  CALL MPPDB_CHECK( PAF, "FLAT_INV beg:PAF" )
  CALL MPPDB_CHECK( PCF, "FLAT_INV beg:PCF" )
  CALL MPPDB_CHECK( PBF, "FLAT_INV beg:PBF" )
  CALL MPPDB_CHECK( PTRIGSX, "FLAT_INV beg:PTRIGSX" )
  CALL MPPDB_CHECK( PTRIGSY, "FLAT_INV beg:PTRIGSY" )
  CALL MPPDB_CHECK( KIFAXX, "FLAT_INV beg:KIFAXX" )
  CALL MPPDB_CHECK( KIFAXY, "FLAT_INV beg:KIFAXY" )
  CALL MPPDB_CHECK( PY, "FLAT_INV beg:PY" )
END IF
!
!*       1.    COMPUTE LOOP BOUNDS
!              -------------------
!
CALL GET_PHYSICAL_ll(IIB,IJB,IIE,IJE)
CALL GET_DIM_EXT_ll('X',IIX,IJX)
CALL GET_DIM_EXT_ll('Y',IIY,IJY)
IIMAX = IIX-2*JPHEXT
IJMAX = IJY-2*JPHEXT
!
IKU=SIZE(PY,3)
IKB=1+JPVEXT
IKE=IKU - JPVEXT
IKMAX=IKE-IKB+1
!
#ifndef MNH_OPENACC
ALLOCATE( ZY (SIZE( PY, 1 ),  SIZE( PY, 2 ),  SIZE( PY, 3 ) ) )
ALLOCATE( ZAF(SIZE( PBF, 1 ), SIZE( PBF, 2 ), SIZE( PBF, 3 )) )

ALLOCATE(ZBAND_X(IIX,IJX,IKU))
ALLOCATE(ZBAND_Y(IIY,IJY,IKU))
ALLOCATE(ZBAND_YR(IIY,IJY,IKU))
ALLOCATE(ZWORKX(IIX,IJX,IKU))
ALLOCATE(ZWORKY(IIY,IJY,IKU))
ALLOCATE(ZBETX(IIY,IJY))
ALLOCATE(ZGAM(IIY,IJY,IKU))
IF (.NOT. L2D) THEN
  ALLOCATE(ZBAND_YT(IJY,IIY,IKU))
  ALLOCATE(ZBAND_YRT(IJY,IIY,IKU))
END IF
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'FLAT_INV' )

CALL MNH_MEM_GET( ZY,  SIZE( PY, 1 ),  SIZE( PY, 2 ),  SIZE( PY, 3 )  )
CALL MNH_MEM_GET( ZAF, SIZE( PBF, 1 ), SIZE( PBF, 2 ), SIZE( PBF, 3 ) )

CALL MNH_MEM_GET( ZBAND_X,  IIX, IJX, IKU )
CALL MNH_MEM_GET( ZBAND_Y,  IIY, IJY, IKU )
CALL MNH_MEM_GET( ZBAND_YR, IIY, IJY, IKU )
CALL MNH_MEM_GET( ZWORKX,   IIX, IJX, IKU )
CALL MNH_MEM_GET( ZWORKY,   IIY, IJY, IKU )
CALL MNH_MEM_GET( ZBETX,    IIY, IJY )
CALL MNH_MEM_GET( ZGAM,     IIY, IJY, IKU )
IF ( .NOT. L2D ) THEN
  CALL MNH_MEM_GET( ZBAND_YT,  IJY, IIY, IKU )
  CALL MNH_MEM_GET( ZBAND_YRT, IJY, IIY, IKU )
ELSE
  CALL MNH_MEM_GET( ZBAND_YT,  0, 0, 0 )
  CALL MNH_MEM_GET( ZBAND_YRT, 0, 0, 0 )
END IF
#endif
!
!-------------------------------------------------------------------------------
!
!*       2.    COMPUTE THE ARRAY INCREMENTS FOR THE FFT
!              ----------------------------------------
!

!$acc data present( PRHOM, PAF, PCF, PBF, PTRIGSX, PTRIGSY, KIFAXX, KIFAXY, PY, PF_1_Y ) &
!$acc &    present( ZY, ZAF, ZBAND_X, ZBAND_Y, ZBAND_YR, ZWORKX, ZWORKY, ZBETX, ZGAM, ZBAND_YT, ZBAND_YRT )

IF(.NOT. L2D) THEN
!
  ILOTX = IJX*IKU
  INC2X = IIX
!
  ILOTY = IIY*IKU
  INC2Y = IJY
!
ELSE
!
  ILOTX = IKU
  INC2X = IIX*IJX
ENDIF
!
!-------------------------------------------------------------------------------
!
!*      3.    FORM HOMOGENEOUS BOUNDARY CONDITIONS FOR A NONCYCLIC CASE
!             ---------------------------------------------------------
!
!
!*      3.1 copy the RHS in a local array REMAP functions will shift the indices for the FFT
!
!$acc kernels
PF_1_Y(:,:,:) = 0.
ZY(:,:,:) = PY(:,:,:)
!$acc end kernels
!
!*      3.2 form homogeneous boundary condition used by the FFT for non-periodic
!           cases
!
!  modify the RHS in the x direction
!
IF (HLBCX(1) /= 'CYCL') THEN
!
  IF (LWEST_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JK=IKB,IKE
      DO JJ = IJB, IJE
        ZY(IIB,JJ,JK)     = ZY(IIB,JJ,JK)     + PY(IIB-1,JJ,JK)
      END DO
    END DO
!$acc end kernels
  END IF
!
  IF (LEAST_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JK=IKB,IKE
      DO JJ = IJB, IJE
        ZY(IIE,JJ,JK)     = ZY(IIE,JJ,JK)     - PY(IIE+1,JJ,JK)
      END DO
    END DO
!$acc end kernels
  END IF
END IF
!
! modify the RHS in the same way along y
!
IF (HLBCY(1) /= 'CYCL'.AND. (.NOT. L2D)) THEN
  IF (LSOUTH_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JK=IKB,IKE
      DO JI = IIB, IIE
        ZY(JI,IJB,JK)     = ZY(JI,IJB,JK)     + PY(JI,IJB-1,JK)
      END DO
    END DO
!$acc end kernels
  END IF
!
  IF (LNORTH_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JK=IKB,IKE
      DO JI = IIB, IIE
        ZY(JI,IJE,JK)     = ZY(JI,IJE,JK)     - PY(JI,IJE+1,JK)
      END DO
    END DO
!$acc end kernels
  END IF
END IF
!
!
!*      3.3 2way structure -> xslice structure, + data shift
!
!$acc kernels
ZBAND_X(:,:,:) = 0.
!$acc end kernels
CALL REMAP_2WAY_X_ll(ZY,ZBAND_X,IINFO_ll)
!
!  
!-------------------------------------------------------------------------------
!
!*      4.    APPLY A REAL TO COMPLEX FFT
!             ---------------------------
!
!
IF (HLBCX(1) == 'CYCL') THEN
    ZBAND1D( 1 : SIZE(ZBAND_X,1)*SIZE(ZBAND_X,2)*SIZE(ZBAND_X(:,:,IKB-1:),3) ) => ZBAND_X(:,:,IKB-1:)
    ZWORK1D( 1 : SIZE(ZWORKX) ) => ZWORKX(:,:,:)
    CALL FFT991( ZBAND1D, ZWORK1D, PTRIGSX, KIFAXX, INC2X, IIMAX, ILOTX, -1 )
ELSE
    CALL FFT55( ZBAND_X(1:,1:,IKB-1:), ZWORKX, PTRIGSX, KIFAXX, -1 )
END IF
!
!
!$acc kernels
ZBAND_Y(:,:,:) = 0.
!$acc end kernels
CALL REMAP_X_Y_ll(ZBAND_X,ZBAND_Y,IINFO_ll)
!
IF (.NOT. L2D) THEN                                 
!
! array transposition I --> J
!
  CALL FAST_TRANSPOSE(ZBAND_Y,ZBAND_YT,IIY,IJY,IKU)
!
  IF (HLBCY(1) == 'CYCL') THEN
       ZBAND1D( 1 : SIZE(ZBAND_YT,1)*SIZE(ZBAND_YT,2)*SIZE(ZBAND_YT(:,:,IKB-1:),3) ) => ZBAND_YT(:,:,IKB-1:)
       ZWORK1D( 1 : SIZE(ZWORKY) ) => ZWORKY(:,:,:)
       CALL FFT991( ZBAND1D, ZWORK1D, PTRIGSY, KIFAXY, INC2Y, IJMAX, ILOTY, -1 )
  ELSE
       CALL FFT55( ZBAND_YT(1:,1:,IKB-1:), ZWORKY, PTRIGSY, KIFAXY, -1 )
  END IF
!
END IF                                        
!
!       singular matrix case : the last term is computed by setting the 
!       average of the pressure field equal to zero.
IF (LWEST_ll(HSPLITTING='Y')) THEN
  IF (L2D) THEN
!$acc kernels
    ZBAND_Y(1,1,IKE+1)=0
!$acc end kernels
  ELSE
!$acc kernels
    ZBAND_YT(1,1,IKE+1)=0.
!$acc end kernels
  END IF
END IF
!
!-------------------------------------------------------------------------------
!
!*      5.    MATRIX INVERSION FOR THE FLAT OPERATOR
!             --------------------------------------
!
CALL FAST_SPREAD(PAF,ZAF,IIY,IJY,IKU)
!
IF (LWEST_ll(HSPLITTING='Y')) THEN
!$acc kernels
   ZAF(1,1,IKE+1)=0. !singular matrix corresponding to the horizontal average
!$acc end kernels
END IF
!
IF (L2D) THEN
#ifdef MNH_OPENACC
CALL PRINT_MSG( NVERB_WARNING, 'GEN', 'FLAT_INV', 'L2D=T not tested' )
#endif
  CALL FAST_SUBSTITUTION_2D(ZBAND_YR,ZBETX,PBF,ZGAM,PCF,ZAF &
                           ,ZBAND_Y,IIY,IJY,IKU)
ELSE
  CALL FAST_SUBSTITUTION_3D(ZBAND_YRT,ZBETX,PBF,ZGAM,PCF,ZAF &
                      ,ZBAND_YT,IIY,IJY,IKU)
END IF
!  
!
!-------------------------------------------------------------------------------
!
!*      6.    APPLY A COMPLEX TO REAL FFT
!             ---------------------------
!
!  
IF (.NOT. L2D) THEN
  IF (HLBCY(1) == 'CYCL') THEN
    ZBAND1D( 1 : SIZE(ZBAND_YRT,1)*SIZE(ZBAND_YRT,2)*SIZE(ZBAND_YRT(:,:,IKB-1:),3) ) => ZBAND_YRT(:,:,IKB-1:)
    ZWORK1D( 1 : SIZE(ZWORKY) ) => ZWORKY(:,:,:)
    CALL FFT991( ZBAND1D, ZWORK1D, PTRIGSY, KIFAXY, INC2Y, IJMAX, ILOTY, +1 )
  ELSE
    CALL FFT55( ZBAND_YRT(1:,1:,IKB-1:), ZWORKY, PTRIGSY, KIFAXY, +1 )
  END IF
  ! array transposition J --> I
  CALL FAST_TRANSPOSE(ZBAND_YRT,ZBAND_YR,IJY,IIY,IKU)
ENDIF
!
! Transposition Y-> X
!
!$acc kernels
ZBAND_X(:,:,:) = 0.
!$acc end kernels
CALL REMAP_Y_X_ll(ZBAND_YR,ZBAND_X,IINFO_ll)
!
!
IF (HLBCX(1) == 'CYCL') THEN
  ZBAND1D( 1 : SIZE(ZBAND_X,1)*SIZE(ZBAND_X,2)*SIZE(ZBAND_X(:,:,IKB-1:),3) ) => ZBAND_X(:,:,IKB-1:)
  ZWORK1D( 1 : SIZE(ZWORKX) ) => ZWORKX(:,:,:)
  CALL FFT991( ZBAND1D, ZWORK1D, PTRIGSX, KIFAXX, INC2X, IIMAX, ILOTX, +1 )
ELSE
  CALL FFT55( ZBAND_X(1:,1:,IKB-1:), ZWORKX, PTRIGSX, KIFAXX, +1 )
END IF
!
!-------------------------------------------------------------------------------
!
!*      7.    RETURN TO A NON HOMOGENEOUS NEUMAN CONDITION FOR NON-CYCLIC CASES
!             -----------------------------------------------------------------
!
!*      7.1   Transposition + shift X -> 2way
!
CALL REMAP_X_2WAY_ll(ZBAND_X,PF_1_Y,IINFO_ll)
!
!*      7.2 complete the lateral boundaries
!
IF (HLBCX(1) /= 'CYCL') THEN
!
!*      7.2.1 return to a non-homogeneous case in the x direction
! 
  ZDXM2 = PDXHATM*PDXHATM
!
  IF (LWEST_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JK=IKB,IKE
      DO JJ = IJB,IJE
        PF_1_Y(IIB-1,JJ,JK) = PF_1_Y(IIB,JJ,JK) - PY(IIB-1,JJ,JK)*ZDXM2/PRHOM(JK)
      END DO
    END DO
!$acc end kernels
  END IF
!
  IF (LEAST_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JK=IKB,IKE
      DO JJ = IJB,IJE
        PF_1_Y(IIE+1,JJ,JK) = PF_1_Y(IIE,JJ,JK) + PY(IIE+1,JJ,JK)*ZDXM2/PRHOM(JK)
      END DO
    END DO
!$acc end kernels
  END IF
!
!    we set the solution at the corner point by the condition:
!    dxm ( P ) = 0
  IF (LWEST_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JJ = IJB,IJE
      PF_1_Y(IIB-1,JJ,IKB-1) = PF_1_Y(IIB,JJ,IKB-1) 
      PF_1_Y(IIB-1,JJ,IKE+1) = PF_1_Y(IIB,JJ,IKE+1)
    END DO
!$acc end kernels
  END IF
  IF (LEAST_ll(HSPLITTING='B')) THEN
!$acc kernels
    DO JJ = IJB,IJE
      PF_1_Y(IIE+1,JJ,IKB-1) = PF_1_Y(IIE,JJ,IKB-1)
      PF_1_Y(IIE+1,JJ,IKE+1) = PF_1_Y(IIE,JJ,IKE+1)
    END DO
!$acc end kernels
  END IF
!
ELSE
!
!*      7.2.2 periodize the pressure function field along the x direction
!
! in fact this part is useless because it is done in the routine
! REMAP_X_2WAY.
!
END IF
!
IF (.NOT.L2D) THEN
  IF (HLBCY(1) /= 'CYCL') THEN
!
!*      7.2.3 return to a non-homogeneous case in the y direction
!
    ZDYM2 = PDYHATM*PDYHATM 
!
    IF (LSOUTH_ll(HSPLITTING='B')) THEN
!$acc kernels
      DO JK=IKB,IKE
        DO JI = IIB,IIE
          PF_1_Y(JI,IJB-1,JK) = PF_1_Y(JI,IJB,JK) - PY(JI,IJB-1,JK)*ZDYM2/PRHOM(JK)
        END DO
      END DO
!$acc end kernels
    END IF
!
    IF (LNORTH_ll(HSPLITTING='B')) THEN
!$acc kernels
      DO JK=IKB,IKE
        DO JI = IIB,IIE
          PF_1_Y(JI,IJE+1,JK) = PF_1_Y(JI,IJE,JK) + PY(JI,IJE+1,JK)*ZDYM2/PRHOM(JK)
        END DO
      END DO
!$acc end kernels
    END IF
!    we set the solution at the corner point by the condition:
!    dym ( P )  = 0
!
    IF (LSOUTH_ll(HSPLITTING='B')) THEN
!$acc kernels
      DO JI = IIB,IIE
        PF_1_Y(JI,IJB-1,IKB-1) = PF_1_Y(JI,IJB,IKB-1)
        PF_1_Y(JI,IJB-1,IKE+1) = PF_1_Y(JI,IJB,IKE+1)
      END DO
!$acc end kernels
    END IF
!
    IF (LNORTH_ll(HSPLITTING='B')) THEN
!$acc kernels
      DO JI = IIB,IIE
        PF_1_Y(JI,IJE+1,IKB-1) = PF_1_Y(JI,IJE,IKB-1)
        PF_1_Y(JI,IJE+1,IKE+1) = PF_1_Y(JI,IJE,IKE+1)
      END DO
!$acc end kernels
    END IF
  ELSE
!
!*      7.2.4 periodize the pressure function field along the y direction
!
!
! in fact this part is useless because it is done in the routine
! REMAP_X_2WAY.
!
  END IF
!
END IF
!
IF (.NOT. L2D .AND. HLBCX(1)/='CYCL' .AND. HLBCY(1)/='CYCL') THEN
! the following verticals are not used
  IF ( (LWEST_ll(HSPLITTING='B')).AND.(LSOUTH_ll(HSPLITTING='B')) ) THEN
!$acc kernels
    PF_1_Y(IIB-1,IJB-1,:)=0.
!$acc end kernels
  END IF
!
  IF ( (LWEST_ll(HSPLITTING='B')).AND.(LNORTH_ll(HSPLITTING='B')) ) THEN
!$acc kernels
    PF_1_Y(IIB-1,IJE+1,:)=0.
!$acc end kernels
  END IF
!
  IF ( (LEAST_ll(HSPLITTING='B')).AND.(LSOUTH_ll(HSPLITTING='B')) ) THEN
!$acc kernels
    PF_1_Y(IIE+1,IJB-1,:)=0.
!$acc end kernels
  END IF
!
  IF ( (LEAST_ll(HSPLITTING='B')).AND.(LNORTH_ll(HSPLITTING='B')) ) THEN
!$acc kernels
    PF_1_Y(IIE+1,IJE+1,:)=0.
!$acc end kernels
  END IF
END IF
!
#ifndef MNH_OPENACC
DEALLOCATE(ZY)
DEALLOCATE(ZAF)
DEALLOCATE(ZBAND_X)
DEALLOCATE(ZBAND_Y)
DEALLOCATE(ZBAND_YT)
DEALLOCATE(ZBAND_YRT)
DEALLOCATE(ZBAND_YR)
DEALLOCATE(ZWORKX)
DEALLOCATE(ZWORKY)
DEALLOCATE(ZBETX)
DEALLOCATE(ZGAM)
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'FLAT_INV' )
#endif

!$acc end data

IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK( PF_1_Y, "FLAT_INV end:PF_1_Y" )
END IF
!
!-------------------------------------------------------------------------------
!
CONTAINS
  SUBROUTINE FAST_TRANSPOSE(PX,PXT,KNI,KNJ,KNK)
    INTEGER,                      INTENT(IN)  :: KNI,KNJ,KNK ! 3D dimension of X and XT
    REAL, DIMENSION(KNI*KNJ,KNK), INTENT(IN)  :: PX
    REAL, DIMENSION(KNJ*KNI,KNK), INTENT(OUT) :: PXT
    !
    INTEGER                      :: IJI,II,IJ,IIJ ! index in array X and XT
    INTEGER                      :: JK
!
!$acc data present( PX, PXT )

!$acc kernels
    DO JK=1,KNK
       ! PERMUTATION(PX,PXT)
       !CDIR NODEP
       !OCL NOVREC
       DO IJI = 1, KNJ*KNI
          ! I,J Indice in XT array from linearised index IJI
          II   = 1 +    (IJI-1)/KNJ
          IJ   = IJI - (II-1)*KNJ 
          ! linearised index in X
          IIJ = II + (IJ-1)*KNI 
          ! transposition
          PXT(IJI,JK) = PX(IIJ,JK)
       
       END DO
    END DO
!$acc end kernels

!$acc end data
!
END SUBROUTINE FAST_TRANSPOSE

SUBROUTINE FAST_SUBSTITUTION_3D(PBAND_YR,PBETX,PPBF,PGAM,PPCF,PAF &
                            ,PBAND_Y,KIY,KJY,KKU)
INTEGER,                      INTENT(IN)  :: KIY,KJY,KKU
REAL, DIMENSION (KIY*KJY,KKU),INTENT(OUT) :: PBAND_YR,PGAM
REAL, DIMENSION (KIY*KJY,KKU),INTENT(IN)  :: PBAND_Y,PPBF,PAF
REAL, DIMENSION (KIY*KJY),    INTENT(OUT) :: PBETX
REAL, DIMENSION (KKU),        INTENT(IN)  :: PPCF
!
INTEGER                        :: JK
!
!
!       initialization 
!
!
!$acc data present( PBAND_YR, PGAM, PBAND_Y, PPBF, PAF, PBETX, PPCF )

!$acc kernels
PBAND_YR = 0.0
PBETX(:) = PPBF(:,IKB-1)
PBAND_YR(:,IKB-1) = PBAND_Y(:,IKB-1)  &
                                      / PBETX(:)
PGAM(:,1:IKB-1) = 0.
!
!        decomposition and forward substitution
!
DO JK = IKB,IKE+1
  PGAM(:,JK) = PPCF(JK-1) / PBETX(:)
!
  PBETX(:) = PPBF(:,JK) -              &      
                 PAF(:,JK)*PGAM(:,JK)
!
  PBAND_YR(:,JK) = ( PBAND_Y(:,JK) -  &
       PAF(:,JK)*PBAND_YR(:,JK- 1) )  &
                                /PBETX(:)
!
END DO
!
!       backsubstitution
!
DO JK = IKE,IKB-1,-1
  PBAND_YR(:,JK) = PBAND_YR(:,JK) -    &
          PGAM(:,JK+1)*PBAND_YR(:,JK+1)
END DO
!$acc end kernels

!$acc end data
!
!
END SUBROUTINE FAST_SUBSTITUTION_3D
!
SUBROUTINE FAST_SUBSTITUTION_2D(PBAND_YR,PBETX,PPBF,PGAM,PPCF,PAF &
                            ,PBAND_Y,KIY,KJY,KKU)
INTEGER,                      INTENT(IN)    :: KIY, KJY, KKU
REAL, DIMENSION(KIY,KJY,KKU), INTENT(OUT)   :: PBAND_YR
REAL, DIMENSION(KIY,KJY,KKU), INTENT(IN)    :: PBAND_Y
REAL, DIMENSION(KIY,KJY,KKU), INTENT(IN)    :: PPBF
REAL, DIMENSION(KIY,KJY,KKU), INTENT(INOUT) :: PGAM
REAL, DIMENSION(KIY,KJY,KKU), INTENT(IN)    :: PAF
REAL, DIMENSION(KIY,KJY),     INTENT(INOUT) :: PBETX
REAL, DIMENSION(KKU),         INTENT(IN)    :: PPCF
!
INTEGER                        :: JK
!
!
!       initialization 
!
!
!$acc data present( PBAND_YR, PGAM, PBAND_Y, PPBF, PAF, PBETX, PPCF )

!$acc kernels
PBAND_YR = 0.0
PBETX(:,1) = PPBF(:,1,IKB-1)
PBAND_YR(:,1,IKB-1) = PBAND_Y(:,1,IKB-1)  &
                                     / PBETX(:,1)
!
!        decomposition and forward substitution
!
DO JK = IKB,IKE+1
  PGAM(:,1,JK) = PPCF(JK-1) / PBETX(:,1)
!
  PBETX(:,1) = PPBF(:,1,JK) -              &      
                 PAF(:,1,JK)*PGAM(:,1,JK)
!
  PBAND_YR(:,1,JK) = ( PBAND_Y(:,1,JK) -  &
       PAF(:,1,JK)*PBAND_YR(:,1,JK- 1) )  &
                                /PBETX(:,1)
!
END DO
!
!       backsubstitution
!
DO JK = IKE,IKB-1,-1
  PBAND_YR(:,1,JK) = PBAND_YR(:,1,JK) -    &
          PGAM(:,1,JK+1)*PBAND_YR(:,1,JK+1)
END DO
!$acc end kernels

!$acc end data
!
!
END SUBROUTINE FAST_SUBSTITUTION_2D

SUBROUTINE FAST_SPREAD(PTAB1D,PTAB3D,KIY,KJY,KKU)
INTEGER,                      INTENT(IN)  :: KIY,KJY,KKU
REAL, DIMENSION(KKU),         INTENT(IN)  :: PTAB1D
REAL, DIMENSION(KIY*KJY,KKU), INTENT(OUT) :: PTAB3D

INTEGER                        :: JIJ,JK
!
!$acc data present( PTAB1D, PTAB3D )

!$acc kernels
DO JK=1,KKU
   DO JIJ=1,KIY*KJY
      PTAB3D(JIJ,JK) = PTAB1D(JK)
   ENDDO
ENDDO
!$acc end kernels

!$acc end data
!
END SUBROUTINE FAST_SPREAD
!
!------------------------------------------------------------------------------  
END SUBROUTINE FLAT_INV
