!MNH_LIC Copyright 1994-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ###################
      MODULE MODI_GRAVITY    
!     ###################
!
INTERFACE
!
      SUBROUTINE GRAVITY    ( KRR,KRRL, KRRI, PTHT, PRT,                   &
                              PRHODJ, PTHVREF, PRWS                        )
!
INTEGER,                  INTENT(IN)    :: KRR  ! Total number of water var.
INTEGER,                  INTENT(IN)    :: KRRL ! Number of liquid water var.
INTEGER,                  INTENT(IN)    :: KRRI ! Number of ice water var.
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PTHT            !     at
REAL, DIMENSION(:,:,:,:), INTENT(INOUT) :: PRT             !  time t
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ    ! Density * Jacobian
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHVREF   ! Virtual Temperature of the reference state
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PRWS      ! Sources of Momentum
!
END SUBROUTINE GRAVITY
!
END INTERFACE
!
END MODULE MODI_GRAVITY 
!     ######################################################################
      SUBROUTINE GRAVITY( KRR,KRRL, KRRI, PTHT, PRT,          &
                          PRHODJ, PTHVREF, PRWS               )
!     ######################################################################
!
!!****  *GRAVITY * - routine to compute the curvature, coriolis and gravity terms
!!
!!    PURPOSE
!!    -------
!!      The purpose of this routine is to compute the gravity on the vertical 
!!    component of the momentum.
!!
!!     
!!**  METHOD
!!    ------
!!
!!    EXTERNAL
!!    --------
!!      MXM,MYM,MZM : Shuman functions (mean operators)
!!      MXF,MYF,MZF : Shuman functions (mean operators)
!!      GZ_M_W      : projection along the vertical direction of the gradient
!!                    vector. It acts on a field localized in mass point and
!!                    the result is localized in w point.
!!      BUDGET      : Stores the different budget components
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      Module MODD_CONF     : contains configuration variables for all models
!!           LTHINSHELL  = .TRUE.  if the THINSHELL approximation is made
!!           LCARTESIAN  = .TRUE.  if the CARTESIAN approximation is made
!!      Module MODD_CST      : 
!!           XG           gravity acceleration
!!           XRADIUS      Earth radius
!!           XRD,XRV      Gas constant for dry air and wator vapor
!!           XCPD,XCPV    Cp for dry air and wator vapor
!!           XCL,XCI      C (calorific capacity) for liquid and solid water
!!      Module MODD_DYN      : contains the parameters for the dynamics
!!           LCORIO      = .FALSE. if the earth rotation is neglected
!!    
!!      Module MODD_BUDGET:
!!         NBUMOD       : model in which budget is calculated
!!         CBUTYPE      : type of desired budget
!!                          'CART' for cartesian box configuration
!!                          'MASK' for budget zone defined by a mask 
!!                          'NONE'  ' for no budget
!!         LBU_RTH      : logical for budget of RTH (potential temperature)
!!                        .TRUE. = budget of RTH        
!!                        .FALSE. = no budget of RTH
!!         LBU_RU       : logical for budget of RU (wind component along x)
!!                        .TRUE. = budget of RU         
!!                        .FALSE. = no budget of RU 
!!         LBU_RV       : logical for budget of RV (wind component along y)
!!                        .TRUE. = budget of RV         
!!                        .FALSE. = no budget of RV 
!!         LBU_RW        : logical for budget of RW (wind component along z)
!!                        .TRUE. = budget of RW         
!!                        .FALSE. = no budget of RW 
!!
!!    REFERENCE
!!    ---------
!!      Book2 of documentation ( routine DYN_SOURCE )
!!
!!    AUTHOR
!!    ------
!!	J.-P. Pinty      * Laboratoire d'Aerologie*
!!
!!    MODIFICATIONS
!!    -------------
!!      C.Lac - March 2011 - Splitted  from dyn_sources
!!      Q.Rodier 06/15 correction on budget
!!      J.Escobar : 15/09/2015 : WENO5 & JPHEXT <> 1
!!      J.L. Redelsperger 03/2021 : Ocean model case
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_CONF
USE MODD_CST
USE MODD_DYN_n, ONLY : LOCEAN

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK, ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
use mode_mppdb

USE MODI_GET_HALO
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN !Kept because some parts are not yet ported with OpenACC
USE MODI_SHUMAN_DEVICE
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                  INTENT(IN)    :: KRR  ! Total number of water var.
INTEGER,                  INTENT(IN)    :: KRRL ! Number of liquid water var.
INTEGER,                  INTENT(IN)    :: KRRI ! Number of ice water var.
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PTHT         !     at
REAL, DIMENSION(:,:,:,:), INTENT(INOUT) :: PRT          !  time t
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ    ! dry Density * Jacobian
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHVREF   ! Virtual Temperature of the reference state
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PRWS      ! Sources of Momentum
!
!
!*       0.2   Declarations of local variables :
!
CHARACTER(LEN=3) :: YNUM
INTEGER          :: IIU, IJU, IKU ! dimensions of dummy arrays
REAL       ::  ZRV_OV_RD    ! = RV / RD
INTEGER    ::  JWATER       ! loop index on the different types of water
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZWORK1, ZWORK2
!
!
!-------------------------------------------------------------------------------

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PRHODJ,   "GRAVITY beg:PRHODJ")
  CALL MPPDB_CHECK(PTHVREF,  "GRAVITY beg:PTHVREF")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PTHT,     "GRAVITY beg:PTHT")
  CALL MPPDB_CHECK(PRT,      "GRAVITY beg:PRT")
END IF

IIU = SIZE( ptht, 1 )
IJU = SIZE( ptht, 2 )
IKU = SIZE( ptht, 3 )

#ifndef MNH_OPENACC
allocate( zwork1(IIU, IJU, IKU ) )
allocate( zwork2(IIU, IJU, IKU ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'GRAVITY' )

CALL MNH_MEM_GET( zwork1, IIU, IJU, IKU )
CALL MNH_MEM_GET( zwork2, IIU, IJU, IKU )
#endif


!$acc data present( PTHT, PRT, PRHODJ, PTHVREF, PRWS, &
!$acc &             zwork1, zwork2 )
!
!*       1.     COMPUTES THE GRAVITY TERM
!	        -------------------------
!
IF( L1D ) THEN     ! no buoyancy for 1D case
!$acc kernels
  PRWS(:,:,:) = 0.
!$acc end kernels
ELSE
!
  IF (LOCEAN) THEN  !ocean case
    CALL GET_HALO(PTHT)
    IF(KRR > 0) THEN
#ifndef MNH_OPENACC
      CALL GET_HALO(PRT(:,:,:,1))
#else
      CALL GET_HALO_D(PRT(:,:,:,1))
#endif
      PRWS(:,:,:) = PRWS + XG * (XALPHAOC*MZM((PTHT - PTHVREF )*PRHODJ) &
                           - XBETAOC*MZM((PRT(:,:,:,1) - XSA00OCEAN)*PRHODJ) )
    ELSE ! unsalted case
      PRWS(:,:,:) = PRWS + XG * XALPHAOC*MZM((PTHT - PTHVREF )*PRHODJ )
    END IF
  ELSE   ! Atmospheric case
    IF( KRR > 0) THEN
!
!   compute the ratio : 1 + total water mass / dry air mass
!
!$acc kernels
      ZRV_OV_RD = XRV / XRD
      ZWORK1(:,:,:) = 1.
!$acc end kernels
      DO JWATER = 1, 1 + KRRL + KRRI
#ifndef MNH_OPENACC
        CALL GET_HALO(PRT(:,:,:,JWATER))
#else
        CALL GET_HALO_D(PRT(:,:,:,JWATER))
#endif
!$acc kernels
        !dir$ concurrent
        ZWORK1(:,:,:) = ZWORK1(:,:,:) + PRT(:,:,:,JWATER)
!$acc end kernels
      END DO
!
!   compute the virtual potential temperature when water is present in any form
!
#ifndef MNH_OPENACC
      CALL GET_HALO(PTHT)
#else
      CALL GET_HALO_D(PTHT)
#endif
!
!$acc kernels
      !dir$ concurrent
      ZWORK2(:,:,:) = PTHT(:,:,:) * (1. + PRT(:,:,:,1)*ZRV_OV_RD) / ZWORK1(:,:,:)
!$acc end kernels
    ELSE
!
!   compute the virtual potential temperature when water is absent
!
!$acc kernels
      ZWORK2(:,:,:) = PTHT(:,:,:)
!$acc end kernels
    END IF
!
!   compute the gravity term
!
#ifndef MNH_OPENACC
    PRWS(:,:,:) = XG * MZM( ( (ZWORK2/PTHVREF) - 1. ) * PRHODJ )
#else
!$acc kernels
    !dir$ concurrent
    ZWORK1(:,:,:) = ( ( ZWORK2(:,:,:) / PTHVREF(:,:,:) ) - 1. ) * PRHODJ(:,:,:)
!$acc end kernels
    CALL MZM_DEVICE(ZWORK1,ZWORK2)
!$acc kernels
    PRWS(:,:,:) = XG * ZWORK2(:,:,:)
!$acc end kernels
#endif
!
!    the extrapolation for the PTHT and the THVREF must be the same at the
!    ground
!
!
  END IF
END IF

!$acc end data

#ifndef MNH_OPENACC
deallocate( zwork1, zwork2 )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'GRAVITY' )
#endif

IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PTHT,     "GRAVITY end:PTHT")
  CALL MPPDB_CHECK(PRT,      "GRAVITY end:PRT")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRWS,     "GRAVITY end:PRWS")
END IF

!-------------------------------------------------------------------------------
!
END SUBROUTINE GRAVITY 
