!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_RRHONG
INTERFACE
SUBROUTINE ICE4_RRHONG(LDSOFT, PCOMPUTE, &
                       &PEXN, PLVFACT, PLSFACT, &
                       &PT,   PRRT, &
                       &PTHT, &
                       &PRRHONG_MR, PB_TH, PB_RR, PB_RG)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PEXN     ! Exner function
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PTHT     ! Theta at t
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRHONG_MR ! Mixing ratio change due to spontaneous freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RG
END SUBROUTINE ICE4_RRHONG
END INTERFACE
END MODULE MODI_ICE4_RRHONG
SUBROUTINE ICE4_RRHONG(LDSOFT, PCOMPUTE, &
                       &PEXN, PLVFACT, PLSFACT, &
                       &PT,   PRRT, &
                       &PTHT, &
                       &PRRHONG_MR, PB_TH, PB_RR, PB_RG)
!!
!!**  PURPOSE
!!    -------
!!      Computes the RRHONG process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XTT
USE MODD_RAIN_ICE_DESCR, ONLY: XRTMIN
USE MODD_PARAM_ICE,      ONLY: LFEEDBACKT
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PEXN     ! Exner function
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PTHT     ! Theta at t
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRHONG_MR ! Mixing ratio change due to spontaneous freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RG
!
!*       0.2  declaration of local variables
!
INTEGER                         :: ISIZE
INTEGER                         :: JL
#ifndef MNH_OPENACC
REAL, DIMENSION(:), allocatable :: ZMASK
#else
REAL, DIMENSION(:), POINTER, CONTIGUOUS :: ZMASK
#endif
!
!$acc data present(PCOMPUTE,PEXN,PLVFACT,PLSFACT,PT,PRRT,PTHT, &
!$acc&                PRRHONG_MR,PB_TH,PB_RR,PB_RG)
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PCOMPUTE,"ICE4_RRHONG beg:PCOMPUTE")
  CALL MPPDB_CHECK(PEXN,"ICE4_RRHONG beg:PEXN")
  CALL MPPDB_CHECK(PLVFACT,"ICE4_RRHONG beg:PLVFACT")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_RRHONG beg:PLSFACT")
  CALL MPPDB_CHECK(PT,"ICE4_RRHONG beg:PT")
  CALL MPPDB_CHECK(PRRT,"ICE4_RRHONG beg:PRRT")
  CALL MPPDB_CHECK(PTHT,"ICE4_RRHONG beg:PTHT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_TH,"ICE4_RRHONG beg:PB_TH")
  CALL MPPDB_CHECK(PB_RR,"ICE4_RRHONG beg:PB_RR")
  CALL MPPDB_CHECK(PB_RG,"ICE4_RRHONG beg:PB_RG")
END IF

isize = Size( prrt )

#ifndef MNH_OPENACC
allocate( zmask(isize) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ICE4_RRHONG' )

CALL MNH_MEM_GET( zmask, isize )

!$acc data present(ZMASK)
#endif
!
!*       3.3     compute the spontaneous freezing source: RRHONG
!
!$acc kernels
PRRHONG_MR(:) = 0.
IF(.NOT. LDSOFT) THEN
  DO JL=1, ISIZE
    ZMASK(JL)=MAX(0., -SIGN(1., PT(JL)-(XTT-35.0))) * & ! PT(:)<XTT-35.0
             &MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JL))) * & ! PRRT(:)>XRTMIN(3)
             &PCOMPUTE(JL)
    PRRHONG_MR(JL)=PRRT(JL) * ZMASK(JL)
  ENDDO
  IF(LFEEDBACKT) THEN
    !Limitation due to -35 crossing of temperature
    DO JL=1, ISIZE
      PRRHONG_MR(JL)=MIN(PRRHONG_MR(JL), MAX(0., ((XTT-35.)/PEXN(JL)-PTHT(JL))/(PLSFACT(JL)-PLVFACT(JL))))
    ENDDO
  ENDIF
ENDIF
DO JL=1, ISIZE
  PB_RG(JL) = PB_RG(JL) + PRRHONG_MR(JL)
  PB_RR(JL) = PB_RR(JL) - PRRHONG_MR(JL)
  PB_TH(JL) = PB_TH(JL) + PRRHONG_MR(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!$acc end kernels

!$acc end data

#ifdef MNH_OPENACC
  !Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
  CALL MNH_MEM_RELEASE( 'ICE4_RRHONG' )
#endif

IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_TH,"ICE4_RRHONG end:PB_TH")
  CALL MPPDB_CHECK(PB_RR,"ICE4_RRHONG end:PB_RR")
  CALL MPPDB_CHECK(PB_RG,"ICE4_RRHONG end:PB_RG")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRRHONG_MR,"ICE4_RRHONG end:PRRHONG_MR")
END IF

!$acc end data

END SUBROUTINE ICE4_RRHONG
