!MNH_LIC Copyright 2005-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    ################ 
     MODULE MODI_TM06_H  
!    ################ 
!
INTERFACE
!
      SUBROUTINE TM06_H(KKB,KKTB,KKTE,PTSTEP,PZZ,PFLXZ,PBL_DEPTH)
!
INTEGER,                INTENT(IN)    :: KKB       ! index of 1st physical level
                                                   ! close to ground 
INTEGER,                INTENT(IN)    :: KKTB      ! first physical level in k
INTEGER,                INTENT(IN)    :: KKTE      ! last physical level in k
REAL,                   INTENT(IN)    :: PTSTEP    ! Double time step
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PZZ       ! altitude of flux levels
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PFLXZ     ! heat flux
REAL, DIMENSION(:,:),   INTENT(INOUT) :: PBL_DEPTH ! boundary layer height
!
!-------------------------------------------------------------------------------
!
END SUBROUTINE TM06_H
!
END INTERFACE
!
END MODULE MODI_TM06_H
!
!     #################################################################
      SUBROUTINE TM06_H(KKB,KKTB,KKTE,PTSTEP,PZZ,PFLXZ,PBL_DEPTH)
!     #################################################################
!
!
!!****  *TM06_H* - computes the Third Order Moments
!!
!!    PURPOSE
!!    -------
!
!!**  METHOD
!!    ------
!!
!!     TOMs are deduced from convective normalized TOMs according to Tomas and
!!     Masson 2006
!!    
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!
!!
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!      V. MAsson and S. Tomas  * Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original         sept. 2005
!!       M.Moge          04/2016 Use openACC directives to port the TURB part of Meso-NH on GPU
!!
!! --------------------------------------------------------------------------
!       
!*      0. DECLARATIONS
!          ------------
!
USE MODD_PARAMETERS, ONLY : XUNDEF

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,  ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
use mode_mppdb
#ifdef MNH_OPENACC
use mode_msg
#endif
!
IMPLICIT NONE
!
!
!*      0.1  declarations of arguments
!
INTEGER,                INTENT(IN)    :: KKB       ! index of 1st physical level
                                                   ! close to ground 
INTEGER,                INTENT(IN)    :: KKTB      ! first physical level in k
INTEGER,                INTENT(IN)    :: KKTE      ! last physical level in k
REAL,                   INTENT(IN)    :: PTSTEP    ! Double time step
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PZZ       ! altitude of flux levels
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PFLXZ     ! heat flux
REAL, DIMENSION(:,:),   INTENT(INOUT) :: PBL_DEPTH ! boundary layer height
!
!-------------------------------------------------------------------------------
!
!       0.2  declaration of local variables
!
!
INTEGER                              :: JK        ! loop counter
#ifndef MNH_OPENACC
LOGICAL, DIMENSION(:,:), allocatable :: GWORK
REAL                                 :: ZGROWTH   ! maximum BL growth rate
REAL,    DIMENSION(:,:), allocatable :: ZFLXZMIN  ! minimum of temperature flux
REAL,    DIMENSION(:,:), allocatable :: ZBL_DEPTH ! BL depth at previous time-step
#else
LOGICAL, DIMENSION(:,:), pointer, contiguous :: GWORK
REAL                                         :: ZGROWTH   ! maximum BL growth rate
REAL,    DIMENSION(:,:), pointer, contiguous :: ZFLXZMIN  ! minimum of temperature flux
REAL,    DIMENSION(:,:), pointer, contiguous :: ZBL_DEPTH ! BL depth at previous time-step
#endif
!----------------------------------------------------------------------------

!$acc data present( PZZ, PFLXZ, PBL_DEPTH )

#ifdef MNH_OPENACC
call Print_msg( NVERB_WARNING, 'GEN', 'TM06_H', 'OpenACC: not yet tested' )
#endif

if ( mppdb_initialized ) then
  !check all in arrays
  call Mppdb_check( pzz,   "Tm06_h beg:pzz"   )
  call Mppdb_check( pflxz, "Tm06_h beg:pflxz" )
  !check all inout arrays
  call Mppdb_check( pbl_depth, "Tm06_h beg:pbl_depth" )
end if

#ifndef MNH_OPENACC
allocate( ZFLXZMIN (size( pzz, 1 ), size( pzz, 2 ) ) )
allocate( ZBL_DEPTH(size( pzz, 1 ), size( pzz, 2 ) ) )
allocate( GWORK    (size( pzz, 1 ), size( pzz, 2 ) ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'TM06_H' )

CALL MNH_MEM_GET( ZFLXZMIN,  size( pzz, 1 ), size( pzz, 2 ) )
CALL MNH_MEM_GET( ZBL_DEPTH, size( pzz, 1 ), size( pzz, 2 ) )
CALL MNH_MEM_GET( GWORK,     size( pzz, 1 ), size( pzz, 2 ) )

!$acc data present( zflxzmin, zbl_depth, gwork )
#endif

!* mixed boundary layer cannot grow more rapidly than 1800m/h
!$acc kernels
ZGROWTH = 2.0 ! (m/s)
!
!----------------------------------------------------------------------------
!
ZBL_DEPTH(:,:) = PBL_DEPTH(:,:)
WHERE(ZBL_DEPTH(:,:)==XUNDEF) ZBL_DEPTH(:,:)=0.
!
PBL_DEPTH(:,:) = XUNDEF
ZFLXZMIN (:,:) = PFLXZ(:,:,KKB)
!$acc end kernels
!
!TODO BUG: paralleliser la boucle aevc OpenACC. Avec pgi14.9/16.07, le DO...WHERE..ENDWHERE...ENDDO ne compile pas
DO JK=KKTB,KKTE
!$acc kernels
  GWORK(:,:) = PFLXZ(:,:,KKB)>0. .AND. PFLXZ(:,:,JK)<ZFLXZMIN(:,:)
  WHERE (GWORK(:,:))
    PBL_DEPTH(:,:) = PZZ  (:,:,JK) - PZZ(:,:,KKB)
    ZFLXZMIN (:,:) = PFLXZ(:,:,JK)
  END WHERE
!$acc end kernels
END DO
!
!$acc kernels
WHERE(PBL_DEPTH(:,:)/=XUNDEF) PBL_DEPTH(:,:)=MIN(PBL_DEPTH(:,:),ZBL_DEPTH(:,:)+ZGROWTH*PTSTEP)
!$acc end kernels

if ( mppdb_initialized ) then
  !Check all inout arrays
  call Mppdb_check( pbl_depth, "Tm06_h end:pbl_depth" )
end if

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'TM06_H' )
#endif

!$acc end data

!----------------------------------------------------------------------------
END SUBROUTINE TM06_H
